
package bg.unisofia.fmi.grosd.ns.impl;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;

import static bg.unisofia.fmi.grosd.ns.enums.NodeServiceTaskType.*;
import bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState;
import bg.unisofia.fmi.grosd.ns.enums.NodeServiceTaskType;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.peristent_task.NSJavaPersistentTask;
import bg.unisofia.fmi.grosd.ns.processor.NSTaskProcessor;
import bg.unisofia.fmi.grosd.ns.processor.impl.NSJavaPersistentTaskInvocationProcessor;
import bg.unisofia.fmi.grosd.ns.processor.impl.NSJavaTaskProcessor;
import bg.unisofia.fmi.grosd.ns.processor.impl.NSJavaPersistentTaskUploadProcessor;
import bg.unisofia.fmi.grosd.ns.remote.NodeService;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskInvocation;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskUpload;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaTask;
import bg.unisofia.fmi.grosd.ns.remote.task.NSTaskRemote;
import bg.unisofia.fmi.grosd.ns.repository.NSRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepository;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.util.NodeServiceUtilities;
import bg.unisofia.fmi.grosd.ns.util.thread.NSTaskCleanerUtil;


/**
 * Acts as a dispatching unit for all operations supported by the Node Service kernel.
 * @see NodeService
 *@author Shenol Yousouf
 */
public class NodeServiceImpl implements NodeService {
  
  // flag is raised only when the service is stopping; if raised, it is never lowered to false again 
  private boolean stopping = false;

  public NodeServiceImpl() {
    NodeServiceInternal.init();
    NSTaskRepository.init();
    loadPersistentTasks();
  }

  public synchronized void uploadTask (NodeServiceTaskType taskType, NSTaskRemote task, NodeServiceClientCallback client) throws RemoteException {

    checkStoppingFlag();
    
    // TODO: log error for every thrown exception
    
    if (taskType == null) {
      throw new RemoteException("Task type is not defined !");
    }
    
    switch (taskType) {
      case NS_JAVA_GENERIC_TASK: handleJavaTask(task, client); break;
      case NS_JAVA_PERSISTENT_TASK_UPLOAD: handleJavaPersistentTaskUpload(task, client); break;
      case NS_JAVA_PERSISTENT_TASK_INVOCATION: handleJavaPersistentTaskInvocation(task, client); break;
      default: throw new RemoteException("Task type " + taskType.toString() + " is unknown or not supported !");
    }
  }  
  
  
  private void handleJavaTask (NSTaskRemote task, NodeServiceClientCallback client) throws RemoteException {
    
    try {
      checkParams(task, client);
    } catch (NodeServiceException e) {
      throw new RemoteException(e.getMessage());
    }
    
   NSJavaTask javaTask = null;
   try {
       javaTask = (NSJavaTask) task;                                                    
   } catch (ClassCastException e) {
     throw new RemoteException("Provided task is not of the declared type - " + NodeServiceTaskType.NS_JAVA_GENERIC_TASK.toString() + ".", e);
   }
   
   NSJavaTaskProcessor processor = new NSJavaTaskProcessor(javaTask, client);
   task.changeActionState(NSTaskActionState.WAITING, "Task with ID " + task.getTaskID() + " is assigned to a Node Service processor, execution is pending.");
   // TODO: in the future, the thread should be acquired from a thread pool
   processor.start();
  }
  
  
  private void handleJavaPersistentTaskUpload (NSTaskRemote task, NodeServiceClientCallback client) throws RemoteException {
    
    try {
      checkParams(task, client);
    } catch (NodeServiceException e) {
      throw new RemoteException(e.getMessage());
    }
    
    NSJavaPersistentTaskUpload persistentTaskUpload = null;
    try {
      persistentTaskUpload = (NSJavaPersistentTaskUpload) task; 
    } catch (ClassCastException e) {
      throw new RemoteException("Provided task is not of the declared type - " + NodeServiceTaskType.NS_JAVA_PERSISTENT_TASK_UPLOAD.toString() + ".", e);
    }
    
    NSJavaPersistentTaskUploadProcessor processor = new NSJavaPersistentTaskUploadProcessor(persistentTaskUpload, client);
    task.changeActionState(NSTaskActionState.WAITING, "Task with ID " + task.getTaskID() + " is assigned to a Node Service processor, execution is pending.");
    // TODO: in the future, the thread should be acquired from a thread pool
    processor.start();
  }


  
  public synchronized void gcTaskResources(String taskID) throws RemoteException {
    
    checkStoppingFlag();
    
    NSTaskRepositoryRecord taskRecord = NSTaskRepository.getTaskRecord(taskID);
    if (taskRecord == null) {
      throw new RemoteException("There is no record of task with ID " + taskID + " on this node.");
    }
    NSTaskActionState actionState = taskRecord.getState();
    if (NodeServiceUtilities.isActionStateFinal(actionState) == false) {
      throw new RemoteException("Task with ID " + taskID + " is in still not in final state; current state is " + actionState); 
    }
    
    // TODO: timeout for deleting a single file can be read from a configuration or a other property instead of being hard-coded
    long timeout = 1000;
    (new NSTaskCleanerUtil(NS_JAVA_GENERIC_TASK, taskID, timeout)).start();
  }
  
  
  public synchronized void startPersistentTask(String keyIdentifier) throws RemoteException {
    
    checkStoppingFlag();
    
    NSRepositoryRecord<NSJavaPersistentTask> record = NSTaskRepository.getServiceRecord(keyIdentifier);
    if (record != null) {
      NSJavaPersistentTask persistentTask = record.getProcessor();
      if (persistentTask.isStopped()) {
        throw new RemoteException("Persistent task for key identifier " + keyIdentifier + " is already started on node.");
      } else {
        persistentTask.setStopped(false);
        return;
      }
    }
    
    // try to load the persistent task if it is not in NS repository
    File persistentTasksFolder = NodeServiceInternal.getNSPersistentTasksFolder();
    if (persistentTasksFolder.exists() == false || persistentTasksFolder.isDirectory() == false || persistentTasksFolder.canRead() == false) {
      throw new RemoteException("Configuration files of persistent tasks on node are either missing or not accessible.");
    }
    
    File persistentTaskFile = NodeServiceUtilities.getFileForPersistentTask(keyIdentifier);
    if (persistentTaskFile.exists() == false || persistentTaskFile.isFile() == false || persistentTaskFile.canRead() == false) {
      throw new RemoteException("Configuration file for persistent task with key identifier " + keyIdentifier + " is either missing or not accessible.");
    }
    
    ObjectInputStream in = null;
    try { 
      in = new ObjectInputStream(new FileInputStream(persistentTaskFile));
      NSJavaPersistentTask persistentTask = (NSJavaPersistentTask) in.readObject();
      
      // no class loading is required because this has been done already during Node Servcie start or during persistent task upload
      // if the logic is changed, switch the classloading flag to true for safety reasons. although this would have a performance impact
      NSTaskRepository.registerService(persistentTask.getKeyID(), persistentTask, false);
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();
      throw new RemoteException("An error occurred while loading persistent task with key identifier " + keyIdentifier, e);
    } finally {
      if (in != null) {
        try {
          in.close();
        } catch (IOException e) {
          // TODO: log error
          e.printStackTrace();
        }
      }
    }  
  }
  
  
  public synchronized void stopPersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException {
    
    checkStoppingFlag();
    
    NSRepositoryRecord<NSJavaPersistentTask> record = NSTaskRepository.getServiceRecord(keyIdentifier);
    
    if (record == null) {
      throw new RemoteException("Persistent task for key identifier " + keyIdentifier + " is already stopped or not registered on node.");
    }
    
    final NSJavaPersistentTask persistentTask = record.getProcessor();
    if (synchronous) {
      persistentTask.stopService(forcedMode);
    } else {
      (new Thread() {
        public void run() {
          persistentTask.stopService(forcedMode);
        }
      }).start();
    }
  }

  
  public synchronized void removePersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException {
    
    checkStoppingFlag();
    
    if (synchronous) {
      removePersistentTaskSynchronously(keyIdentifier, forcedMode);
    } else {
      (new Thread() {
        public void run() {
          try {
            removePersistentTaskSynchronously(keyIdentifier, forcedMode);
          } catch (RemoteException e) {
            // TODO: log error in asynchronous mode
            e.printStackTrace();
          }
        }
      }).start();
    }
    
  }
  
  
  private void removePersistentTaskSynchronously(String keyIdentifier, final boolean forcedMode) throws RemoteException {

    // step 1: stop persistent task if it is loaded in repository
    // step 2: remove its files from Node Service file system
    
    NSRepositoryRecord<NSJavaPersistentTask> record = NSTaskRepository.getServiceRecord(keyIdentifier);

    long timeout = 1000;
    
    if (record != null) {
      NSJavaPersistentTask persistentTask = record.getProcessor();
      persistentTask.stopService(forcedMode);
      
      try {
        // delete persistent task's files in a synchronous call
        (new NSTaskCleanerUtil(NS_JAVA_PERSISTENT_TASK_UPLOAD, keyIdentifier, timeout)).run();
      } catch (NodeServiceException e) {
        // TODO: log error
        e.printStackTrace();
        throw new RemoteException(e.getMessage());
      }
      
      NSTaskRepository.removeServiceRecord(keyIdentifier);
    }
    
    // now that the persistent task is no longer in NS repository, remove its file. so that it not loaded on next start of the Node Service
    File persistantTaskFile = NodeServiceUtilities.getFileForPersistentTask(keyIdentifier);
    if (persistantTaskFile != null && persistantTaskFile.exists()) {
      NodeServiceUtilities.deleteFileWithTimeout(persistantTaskFile, timeout);
    }    
  }
  

  public synchronized void stopJavaTask(String taskID, final boolean synchronous) throws RemoteException {
    
    checkStoppingFlag();
    
    final NSTaskRepositoryRecord record = NSTaskRepository.getTaskRecord(taskID);
    if (record == null) {
      throw new RemoteException("There is no task with ID " + taskID + " on node.");
    }
    
    if (synchronous) {
      try {
        record.getProcessor().stopTask(true);
      } catch (Exception e) {
        throw new RemoteException(e.getMessage());
      }
    } else {
      (new Thread() {
        public void run() {
          try {
            record.getProcessor().stopTask(false);
          } catch (NodeServiceException e) {
            // TODO: log error in asynchronous mode
            e.printStackTrace();
          }
        }
      }).start();
    }
  }

  
  public synchronized void stopNodeService(final boolean forcedMode, final boolean synchronous) throws RemoteException {
    
    checkStoppingFlag();
    
    if (synchronous) {
      stopNodeServiceSynchronously(forcedMode);
    } else {
      (new Thread() {
        public void run() {
          stopNodeServiceSynchronously(forcedMode);
        }
      }).start();
    }
  }
  
  private void stopNodeServiceSynchronously (boolean forcedMode) {
    
    // step 1: stop all services
    // step 2: stop all other tasks
    // step 3: stop VM
    
    ArrayList<NSJavaPersistentTask> persistentTasks = NSTaskRepository.getAllRegisteredPersistentTasks();
    for (NSJavaPersistentTask task : persistentTasks) {
      try {
        task.stopService(forcedMode);
      } catch (Exception e) {
        // at current implementation, stoping of one processor should not disturb Node Service's stopping procedure
        e.printStackTrace();
      }
    }
    
    ArrayList<NSTaskProcessor> tasks = NSTaskRepository.getAllRegisteredTaskProcessors();
    for (NSTaskProcessor processor : tasks) {
      try {
        processor.stopTask(true);
        //  delete  task's files in a synchronous call
        long timeout = 1000;
        (new NSTaskCleanerUtil(NS_JAVA_GENERIC_TASK, processor.getTaskID(), timeout)).run();
      } catch (Exception e) {
        // at current implementation, stoping of one processor should not disturb Node Service's stopping procedure
        e.printStackTrace();
      }
    }
    
    // at this point, it is assumed that all Node Service Processors are stopped
    // TODO: it should be tested if this exit call stops the Node Service Server too, as intended
    System.exit(0);
  }
  
  
  private void checkParams(NSTaskRemote task, NodeServiceClientCallback client) throws NodeServiceException{
    
    if (task == null) {
      throw new NodeServiceException("No task provoded for upload");
    }
    
    if (client == null) {
      throw new NodeServiceException("No callback connection provided - task cannot be performed.");
    }
  }
  
  private void loadPersistentTasks() {
    File persistentTasksFolder = NodeServiceInternal.getNSPersistentTasksFolder();
    if (persistentTasksFolder.exists() == false || persistentTasksFolder.isDirectory() == false || persistentTasksFolder.canRead() == false) {
      return;
    }
    
    File[] persistentTasksFiles = persistentTasksFolder.listFiles(new FileFilter () {
      public boolean accept(File pathname) {
        return (pathname.isFile() && pathname.getName().endsWith(NodeServiceUtilities.SERIALIZABLE_EXTENTION));
      }
    });
    
    if (persistentTasksFiles != null && persistentTasksFiles.length > 0) {
      for (File file : persistentTasksFiles) {
        ObjectInputStream in = null;
        try { 
          in = new ObjectInputStream(new FileInputStream(file));
          NSJavaPersistentTask persistentTask = (NSJavaPersistentTask) in.readObject();
          
          // class loading is required on first persistent task start
          NSTaskRepository.registerService(persistentTask.getKeyID(), persistentTask, true);
        } catch (Exception e) {
          // currently, any errors in loading a persistent task are not treated as fatal
          // for starting the Node Service; in a  more strict design, however, probably it is  more correct
          // to prevent the Node Service from starting at all
          // TODO: log error
          e.printStackTrace();
        } finally {
          if (in != null) {
            try {
              in.close();
            } catch (IOException e) {
              // TODO: log error
              e.printStackTrace();
            }
          }
        }
      }
    }
  }
  
  
  private void handleJavaPersistentTaskInvocation (NSTaskRemote task, NodeServiceClientCallback client) throws RemoteException {
      try {
        checkParams(task, client);
      } catch (NodeServiceException e) {
        throw new RemoteException(e.getMessage());
      }
      
      NSJavaPersistentTaskInvocation invocationTask = null;
      try {
        invocationTask = (NSJavaPersistentTaskInvocation) task;
      } catch (ClassCastException e) {
        throw new RemoteException("Provided task is not of the declared type - " + NodeServiceTaskType.NS_JAVA_PERSISTENT_TASK_INVOCATION.toString() + ".", e);
      }
      
      NSJavaPersistentTaskInvocationProcessor processor = new NSJavaPersistentTaskInvocationProcessor(invocationTask, client);
      task.changeActionState(NSTaskActionState.WAITING, "Task with ID " + task.getTaskID() + " is assigned to a Node Service processor, execution is pending.\r\n" +
              "Note that the execution could be substantially delayed if the persistent task is already busy with a previous request.");
      // TODO: in the future, the thread should be acquired from a thread pool
      processor.start();
    }
  
  private void checkStoppingFlag() throws RemoteException {
    if (stopping) {
      throw new RemoteException("Node Service is in stopping state, so no further requests can be executed.");
    }
  }
    
}
  
  

