
package bg.unisofia.fmi.grosd.ns.peristent_task.impl;

import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.ABORTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.FINISHED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STARTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STOPPED;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashSet;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.application.StoppableApplication;
import bg.unisofia.fmi.grosd.ns.classloader.NodeServiceClassLoader;
import bg.unisofia.fmi.grosd.ns.exception.NSClassLoaderException;
import bg.unisofia.fmi.grosd.ns.exception.NSJavaPesristentTaskException;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.peristent_task.NSJavaPersistentTask;
import bg.unisofia.fmi.grosd.ns.peristent_task.impl.serializables.NSTaskMethodPrototypeSerializable;
import bg.unisofia.fmi.grosd.ns.peristent_task.impl.serializables.NSTaskTargetClassSerializable;
import bg.unisofia.fmi.grosd.ns.processor.NSTaskProcessor;
import bg.unisofia.fmi.grosd.ns.processor.visitor.NSMethodArgumentVisitor;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskInvocation;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.NSTaskArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.visitor.NSTaskArgumentVisitor;
import bg.unisofia.fmi.grosd.ns.remote.task.method.NSTaskMethodPrototype;
import bg.unisofia.fmi.grosd.ns.remote.task.result.locator.NSFileResultLocator;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepository;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.result.NSTaskResultImpl;
import bg.unisofia.fmi.grosd.ns.util.NSClassChecker;
import bg.unisofia.fmi.grosd.ns.util.NSFileTranferUtilities;
import bg.unisofia.fmi.grosd.ns.util.NSTaskLifecycleNotifier;
import bg.unisofia.fmi.grosd.ns.util.NodeServiceUtilities;


/**
 * @see NSJavaPersistentTask
 *@author Shenol Yousouf
 */
public class NSJavaPersistentTaskImpl implements NSJavaPersistentTask {
  
  private String keyID;
  private boolean busy;
  
  private File[] binaries;
  private NSTaskMethodPrototypeSerializable method;
  private NSTaskTargetClassSerializable targetClass;
  
  // the rest of the fields are NOT for serialization
  
  // these are the cached objects which are target for method calls
  private Object instance;
  private Method targetMethod;
  
  private NSJavaPersistentTaskInvocation invocationTask;
  private NSTaskRepositoryRecord taskRecord;
  private String taskID;
  private NodeServiceClientCallback client;
  private TaskExecutionThread taskThread;
  
  private File processingFolder = new File(System.getProperty("user.dir"));
  
  // this flag is raised only when the task execution should be stopped
  private boolean stopping = false;
  
  // this flag is raised only when the persistent task is stopped and no more requests for execution are accepted
  private boolean stopped = false;
  
  
  NSJavaPersistentTaskImpl (String keyID, File[] binaries, NSTaskMethodPrototypeSerializable method, NSTaskTargetClassSerializable targetCass) {
    this.keyID = keyID;
    this.binaries = binaries;
    this.method = method;
    this.targetClass = targetCass;
    this.busy = false;
  }

  public String getKeyID() {
    return keyID;
  }
  
  public String getCurrentTaskID() {
    return taskID;
  }

  public synchronized boolean isBusy() {
    return busy;
  }

  public void process(NSJavaPersistentTaskInvocation arg, NSTaskProcessor processor, NodeServiceClientCallback nsclient) {

    lock();
    
    invocationTask = arg;
    client = nsclient;
    
    final Set<File> filesForGC = new HashSet<File>();
    
    final String targetClassname = targetClass.getClassName();
    final String methodName = method.getMethodName();
    
    try {
      // check key
      if (invocationTask.getKeyIdentifier().equals(keyID) == false) {
        throw new NSJavaPesristentTaskException("Mismatch in task key - key of persistent task is " +  keyID + ", key of invocation argument is " + invocationTask.getKeyIdentifier());
      }
      
      taskID = invocationTask.getTaskID();      
      taskRecord = NSTaskRepository.registerTask(taskID, processor);
      
      if (isStopped()) {
        throw new NSJavaPesristentTaskException("The persistent task is stopped on node and cannot handle any requests now. Start the persistent task again.");
      }
       
       // step 1: transfer additional files and process arguments for invocation
       // step 2: prepare method for invocation
       // step 3: prepare target class instance if method is not static
       // step 4: execute method; results are collected and returned in a separate thread
       
       // step 1/4
       
       NSFileTranferUtilities.transferClientFiles(filesForGC, client, processingFolder, invocationTask.getAdditionalFiles(), false);
       
       NSTaskArgument[] arguments = invocationTask.getArguments();
       NSMethodArgumentVisitor visitorImpl = new NSMethodArgumentVisitor(client, filesForGC, processingFolder, (arguments != null) ? arguments.length : 0, true);
       if (arguments != null) {
         NSTaskArgumentVisitor visitor = (NSTaskArgumentVisitor) UnicastRemoteObject.exportObject(visitorImpl, 0);
         for(NSTaskArgument argument : arguments) {
           argument.accept(visitor);
         }
       }
       
       // step 2/4
       
       Method currentMethod = null;
       
       // check for custom method
       NSTaskMethodPrototype customMethodPrototype = invocationTask.getMethodPrototype();
       if (customMethodPrototype == null) {
         // use default method
         if (targetMethod == null) {    // target method is not cached
           targetMethod = NSClassChecker.parseMethodSignature(targetClassname, methodName, method.getParameterClassnames());           
         }
         currentMethod = targetMethod;
       } else {
         Class[] parameterTypes = null;
         
         String[] parameterClassnames = customMethodPrototype.getParameterClassnames();
         if (parameterClassnames != null && parameterClassnames.length > 0) {
           parameterTypes = new Class[parameterClassnames.length];
           byte index = 0;
           Class parameterClass = null;
           for(String parameterClassname : parameterClassnames) {
             try {
               parameterClass = NSClassChecker.parseClassname(parameterClassname);
             } catch (Exception e) {
               throw new NSJavaPesristentTaskException("TaskID " + taskID + ": Error loading parameter class for name " + parameterClassname, e);
             }
             parameterTypes[index++] = parameterClass;  
           }
         }
         
         if (parameterTypes == null) {   // read types from the argument classes
           parameterTypes = visitorImpl.getClasses();
          }
         
         currentMethod = NSClassChecker.parseMethodSignature(targetClassname, customMethodPrototype.getMethodName(),  parameterTypes);
       }
       
       // step 3/4
       
       Object currentInstance = null;
       
       if (Modifier.isStatic(currentMethod.getModifiers()) == false) {

         // check for custom initialization
         String[] consParameterTypeNames = invocationTask.getInitParameterClassnames();
         NSTaskArgument[] initArgs = invocationTask.getInitArgs();
         
         if (initArgs == null) {
           // no custom initialization, use default instance
           if (instance == null) {  // instance is not cached
             instance = targetClass.buildNewInstance();
           }
           currentInstance = instance;
         } else if (initArgs.length == 0) {
           // custom invocation of target class default constructor
           Class c = Class.forName(targetClassname, true, NodeServiceInternal.getClassLoader());
           currentInstance = c.newInstance();
         } else {
           
           Class[] consParameterTypes = null;
           byte index = 0;
           Class consParameterClass = null;
           if (consParameterTypeNames != null && consParameterTypeNames.length > 0) {
             consParameterTypes = new Class[consParameterTypeNames.length];
             for (String consParameterTypesName : consParameterTypeNames) {
               try {
                 consParameterClass = Class.forName(consParameterTypesName, true, NodeServiceInternal.getClassLoader());
               } catch (Exception e) {
                 throw new NSJavaPesristentTaskException("TaskID " + taskID + ": Error loading constructor parameter class for name " + consParameterTypesName, e);
               }
               consParameterTypes[index++] = consParameterClass;
             }
           }
           
           NSMethodArgumentVisitor consVisitorImpl = new NSMethodArgumentVisitor(client, filesForGC, processingFolder, initArgs.length, (consParameterTypes == null));
           NSTaskArgumentVisitor consVisitor = (NSTaskArgumentVisitor) UnicastRemoteObject.exportObject(consVisitorImpl, 0);
            for(NSTaskArgument initArg:initArgs) {
              initArg.accept(consVisitor);
            }
            
            if (consParameterTypes == null) {      // read it from argument types
              consParameterTypes = consVisitorImpl.getClasses();
            }
            
            Constructor cons = NSClassChecker.parseConsSignature(targetClassname, consParameterTypes);
            currentInstance = cons.newInstance(consVisitorImpl.getArgValues());
         }
       }       
       
       // step 4/4
       
       taskThread = new TaskExecutionThread(currentMethod, currentInstance, visitorImpl.getArgValues());
       taskThread.start();       
       
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();      
      notifyTaskAborted(e);
    } finally {
      
      if (taskRecord != null && filesForGC.size() > 0) {
        taskRecord.addFilesForGC(filesForGC.toArray(new File[] {}));
      }
      
    }
  }

  public synchronized void stopTask(boolean synchronous) {
    if (taskRecord.isActionStateFinal()) {
      // task is already stopped
      return;
    }
    
    if (taskRecord.getState() == STARTED) {
      // task execution is already in progress and task thread is delegated to stop it
      taskThread.stopTask();
      notifyTaskStopped();
    } else {
      // execution has not bugun yet;
      // the next lock on the current processor will be acquired by the task thread when calling notifyTaskStarted();
      // for this purpose, the stopping flag is raised to signal that the execution should be terminated
      stopping = true;
      
      if (synchronous) {
        // this block intends to make the call of stopTask synchronous with the actual task termination,
        // i.e. the method will return only after the task is really stopped
        while(taskRecord.isActionStateFinal() == false) {
          try {
            wait();
          } catch (InterruptedException e) {
            // TODO: log  error
            e.printStackTrace();
          }
        }
      }
    }
  }
  
  public synchronized void stopService(boolean forcedMode) {
    
    if (isStopped()) {
      // nothing more to do
      return;
    }
    
    if (isBusy()) {
      if (forcedMode == false) {
        // we are in the middle of executing a request, wait for end of execution
        while(isBusy()) {
          try {
            wait();
          } catch (InterruptedException e) {
            // TODO: log error
            e.printStackTrace();
          }
        }
      } else {
        // stop the execution
        stopTask(true);
      }
    }
    
    // the persistent task is not in busy state now;
    // np physical stop of the persistent task is actually possible,
    // so simply raise the flag for stopped state
    setStopped(true);
  }
  
  void registerFilesForGC(File...files) {
    taskRecord.addFilesForGC(files);
  }
  
  // custom serialization behaviour
  private void writeObject (ObjectOutputStream out) throws IOException {
    out.writeUTF(keyID);
    out.writeObject(binaries);
    out.writeObject(method);
    out.writeObject(targetClass);
  }
  
  // custom deserialization behaviour
  private void readObject (ObjectInputStream in) throws IOException, ClassNotFoundException {
    
    this.keyID = in.readUTF();
    this.binaries = (File[])in.readObject();
    this.method = (NSTaskMethodPrototypeSerializable) in.readObject();
    this.targetClass = (NSTaskTargetClassSerializable) in.readObject();
    this.processingFolder = new File(System.getProperty("user.dir"));
    this.busy = false;
  }
  
  synchronized void notifyTaskStarted() throws NodeServiceException {
    
    // this is the only method for notifying that the task action state has changed to a non-final one;
    // for this purpose, it is necessary to check if there is a stop request only here.
    if (stopping) {
      taskThread.stopTask();
      notifyTaskStopped();
      return;
    }
    
    taskRecord.setState(STARTED);
    try {
     invocationTask.changeActionState(STARTED, "Task with ID " + taskID + " successfully started.");
    } catch (RemoteException e) {
      // TODO: log warning / error; no need to abort task for failing to change its state
      e.printStackTrace();
    }    
  }
  
  synchronized void notifyTaskFinished(Serializable result) throws NodeServiceException {
    
    NSFileResultLocator fileResultLocator = null;
    File[] fileResults = null;
    NSTaskResultImpl taskResult = null;
    StringBuilder resultComment = new StringBuilder();
    
    try {
      try {
        fileResultLocator = invocationTask.getFileResultLocator();
      } catch (RemoteException e) {
        // TODO: log error; an attempt will be made to transfer at least the result
        e.printStackTrace();
        resultComment.append("Error in retrieving locator for file results\nCaused by: " + NodeServiceUtilities.formatExceptionMessage(e));
      }
      
      if (fileResultLocator != null) {
        try {
          StringBuilder searchComments = new StringBuilder();
          fileResults = NodeServiceUtilities.getOutputFiles(processingFolder, fileResultLocator, searchComments);
          resultComment.append(searchComments.toString());
        } catch (Exception e) {
          resultComment.append("Error in getting file results.\nCaused by: " + NodeServiceUtilities.formatExceptionMessage(e));
        }
      }
      
     
      if (fileResults == null) {
        if (fileResultLocator != null) {
          resultComment.append("No file results were found ! Possible error in task execution !");
        }
        taskResult = new NSTaskResultImpl(result, processingFolder, new File[] {}, client, resultComment.toString());
      } else {
        registerFilesForGC(fileResults);
        taskResult = new NSTaskResultImpl(result, processingFolder, fileResults, client, resultComment.toString());
      }
      
      taskRecord.setState(FINISHED);
      try {
        invocationTask.changeActionState(FINISHED, "Task with ID " + taskID + " finished successfully.");
        NSTaskLifecycleNotifier.notifyTaskFinished(invocationTask, taskResult);
      } catch (RemoteException e) {
        // TODO: log  error
        e.printStackTrace();
      }
    } finally {
      release();
    }
  }

  synchronized void notifyTaskStopped()  {
    
    try {
      // if task is already in final state, do not propagate the exception to upper layers -
      // task is in stopped state, anyway
      taskRecord.setState(STOPPED);
      // notify any synchronous listeners
      notifyAll();
      invocationTask.changeActionState(STOPPED, "Task with ID " + taskID + " stopped on request.");
    } catch (RemoteException e) {
      // TODO: log  error
      e.printStackTrace();
    } finally {
      release();
    }
  }
  
  synchronized void notifyTaskAborted(Exception ex)  {
    try {
      taskRecord.setState(ABORTED);
    } catch (NodeServiceException e) {
      // TODO: log  error
      e.printStackTrace();
    }
    
    try {
      invocationTask.changeActionState(ABORTED, "Task with ID " + taskID + " has been aborted. Reason: " + NodeServiceUtilities.formatExceptionMessage(ex));
      NSTaskLifecycleNotifier.notifyTaskAborted(invocationTask, "Task with ID " + taskID + "  aborted on start - see exception for cause.", new Exception[] {ex});
    } catch (Exception e) {
      // TODO: log  error
      e.printStackTrace();
    } finally {
      release();
    }
  }
  
  private synchronized void lock() {
    busy = true;
  }
  
  private synchronized void release() {
    busy = false;
    notifyAll();
  }
  
  public synchronized boolean isStopped() {
    return stopped;
  }
  
  public synchronized void setStopped(boolean value) {
    stopped = value;
  }
  
  /**
   * A helper thread for carrying out the actual method invocation, after all the initial metadata
   * has been processed and the method for invocation was constucted.
   * The execution in thread also creates a unit for control of the execution lidecycle itself.
   * @author Shenol Yousouf
   */
  class TaskExecutionThread extends Thread {
    private Method method;
    Object instance;
    private Object[] arguments;
    
    public TaskExecutionThread(Method method, Object instance, Object... arguments) {
      this.method = method;
      this.instance = instance;
      this.arguments = arguments;
    }
    
    public void run() {
      try {
        // this notification is not entirely correct, because we have not invoked the method yet;
        // however, this will prevent confusing situations when client is notified that the task is finished or aborted
        // even before being notified that task was started
        notifyTaskStarted();
      
        Serializable result = null;
        try {
          result = (Serializable) method.invoke(instance, arguments);
        } catch (Exception e) {
          notifyTaskAborted(e);
          return;
        }
        if (result != null && result instanceof File) {
          registerFilesForGC((File)result);
        }
        notifyTaskFinished(result);
      } catch (NodeServiceException e) {
        // TODO: log  error
        notifyTaskAborted(e);
      }
    }
    
    @SuppressWarnings("deprecation")
    public  void stopTask() {
      if (instance instanceof StoppableApplication) {
       ((StoppableApplication) instance).stopExecution();
      }
      // though not elegant, it will ensure stopping this thread
      this.stop();
    }
    
  }

  public void loadServiceClasses() throws NSClassLoaderException {
    URL[] binariesURLs = new URL[binaries.length];
    int index = 0;
    
    // convert bianry files into URLs;
    // the procedure is different for .class and for .jar files
    try {
      for(File binary:binaries) {
        if (binary.getName().toLowerCase().endsWith(NSFileTranferUtilities.JAR_EXTENTION)) {
          binariesURLs[index++] = binary.toURI().toURL();
        } else if (binary.getName().toLowerCase().endsWith(NSFileTranferUtilities.CLASS_EXTENTION)) {
          binariesURLs[index++] = binary.getParentFile().toURI().toURL();
        }
      }
    } catch (MalformedURLException ex) {
      throw new NSClassLoaderException("Error getting URLs from persistent tasks binaries.", ex);
    }
    
    // add all generated URLs to NS classloader
    // TODO: what is the effect if the URLs are already registered with the class loader ?
    NodeServiceClassLoader classLoader = NodeServiceInternal.getClassLoader();
    for(URL binaryURL:binariesURLs) {
      classLoader.addURL(binaryURL);
    }
  }
 
}
