
package bg.unisofia.fmi.grosd.ns.peristent_task.impl.serializables;

import java.io.File;
import java.io.Serializable;
import java.rmi.server.UnicastRemoteObject;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.exception.NSInitializationException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.processor.visitor.NSMethodArgumentVisitor2;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.NSTaskArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.visitor.NSTaskArgumentVisitor;
import bg.unisofia.fmi.grosd.ns.remote.task.method.NSTaskMethodPrototype;
import bg.unisofia.fmi.grosd.ns.remote.task.targetclass.NSTaskTargetClass;
import bg.unisofia.fmi.grosd.ns.util.NSClassChecker;


/**
 * Factroy for contructing the Serializable elements of the persistent task.
 *@author Shenol Yousouf
 */
public abstract class NSJavaPersistentTaskSerializablesFactory {

  /**
   * 
   * @param classname class name from which the method will be loaded
   * @param remote a <code>NSTaskMethodPrototype</code> input object supplied through the API
   * @return a <code>NSTaskMethodPrototypeSerializable</code> instance
   * @throws NSInitializationException
   */
  public static NSTaskMethodPrototypeSerializable buildNSTaskMethodPrototypeSerializable(String classname, NSTaskMethodPrototype remote) throws NSInitializationException {

    try {
      String methodName = remote.getMethodName();
      String[] parameterClassnames = remote.getParameterClassnames();
      
      // check method signature
      NSClassChecker.parseMethodSignature(classname, methodName, parameterClassnames);
      
      return new NSTaskMethodPrototypeSerializable(methodName, parameterClassnames);
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();
      throw new NSInitializationException("Error creating instrance of " + NSTaskMethodPrototypeSerializable.class.getName() + " from remote instance of " + remote.getClass().getName(), e);
    } 
  }
  
  /**
   * 
   * @param remote a <code>NSTaskTargetClass</code> input instance supplied through the API
   * @param client callback connection with the API
   * @param filesForGC list populated with files for removal after executing the persistent task invocation
   * @return a <code>NSTaskTargetClassSerializable</code> instance
   * @throws NSInitializationException
   */
  public static NSTaskTargetClassSerializable buildNSTaskTargetClassSerializable(NSTaskTargetClass remote,  NodeServiceClientCallback client, Set<File> filesForGC) throws NSInitializationException {
    try {
      String classname = remote.getClassName();
      checkClass(classname);
      
      String[] parameterClassnames = remote.getParameterClassnames();
      if (parameterClassnames != null && parameterClassnames.length == 0) {
        parameterClassnames = null;
      }
      
      final File currentFolder = new File (System.getProperty("user.dir"));
      NSTaskArgument[] arguments = remote.getInitArgs();
      NSMethodArgumentVisitor2 visitorImpl = new NSMethodArgumentVisitor2(client, filesForGC, currentFolder, arguments.length, (parameterClassnames == null));
      NSTaskArgumentVisitor visitor = (NSTaskArgumentVisitor) UnicastRemoteObject.exportObject(visitorImpl, 0);
      for(NSTaskArgument argument : arguments) {
        argument.accept(visitor);
      }
      if (parameterClassnames == null) {
        parameterClassnames = visitorImpl.getClassnames();
      }
      Serializable[] initArgs = visitorImpl.getArgValues();
      
      // check initialization parameters
      NSClassChecker.parseConsSignature(classname, parameterClassnames);
      
      return new NSTaskTargetClassSerializable(classname, parameterClassnames, initArgs);
      
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();
      throw new NSInitializationException("Error creating instrance of " + NSTaskTargetClassSerializable.class.getName() + " from remote instance of " + remote.getClass().getName(), e);
    }    
  }
  
  /**
   * 
   * @param classname class name for loading
   * @return if the class with the provided name can be successfully loaded by the NodeServiceClassLoader
   * @throws ClassNotFoundException
   */
  private static Class checkClass(String classname) throws ClassNotFoundException {
    return Class.forName(classname, true, NodeServiceInternal.getClassLoader());
  }
  
  
  
  
}
