
package bg.unisofia.fmi.grosd.ns.peristent_task.impl.serializables;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import bg.unisofia.fmi.grosd.ns.exception.NSInitializationException;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.util.NSClassChecker;


/**
 * Local Serilizable version of <code>NSTaskTargetClass</code>, used for constructing a persistent task.
 *@author Shenol Yousouf
 */
public class NSTaskTargetClassSerializable implements Serializable {

  private String classname;
  private String[] initializationClassnames;
  private Serializable[] initializationArguments;
  
  NSTaskTargetClassSerializable (String classname, String[] initializtionClassnames, Serializable[] initializationArguments) {
    this.classname = classname;
    this.initializationClassnames = initializtionClassnames;
    this.initializationArguments = initializationArguments;
  }
  
  /**
   * 
   * @return target class name
   */
  public String getClassName() {
    return classname;
  }  
  
  /**
   * 
   * @return class names of initialization parameters for the class instance used for the actual task execution;
   * can be <code>null</code> or an empty list because the class names can be restored from the supplied initiazlization values
   */
  public String[] getParameterClassnames() {
    return initializationClassnames;
  }
  
  /**
   * 
   * @return actual initialization arguments for the class instance used for the actual task execution
   */
  public Serializable[] getInitArgs() {
    return initializationArguments;
  }
  
  /**
   * 
   * @return a class instance used for the actual task execution
   * @throws NodeServiceException
   */
  public Object buildNewInstance() throws NodeServiceException {
    Class c;
    Object result = null;
    try {
      c = Class.forName(classname, true, NodeServiceInternal.getClassLoader());
      
      // it is assumed that NSTaskTargetClassSerializable constructor is accessible only for factory methods
      // and thus the list of intialization classnames is consistent with list of argument values
      if (initializationArguments == null || initializationArguments.length == 0) {
        // invoke default constructor
        result = c.newInstance();
      } else {
        Constructor cons = NSClassChecker.parseConsSignature(classname, initializationClassnames);
        result = cons.newInstance(initializationArguments);
      }
    } catch (ClassNotFoundException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NodeServiceException("Cannot get class for classname " + classname, e);
    } catch (InstantiationException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NSInitializationException("Cannot get instance of class " + classname, e);
    } catch (IllegalAccessException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NSInitializationException("Cannot get instance of class " + classname, e);
    } catch (IllegalArgumentException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NSInitializationException("Cannot get instance of class " + classname, e);
    } catch (InvocationTargetException e) {
      e.printStackTrace();
    }
    
    return result;
  }
}
