
package bg.unisofia.fmi.grosd.ns.processor.impl;

import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.ABORTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.FINISHED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STARTED;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.exception.NSClassLoaderException;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.peristent_task.NSJavaPersistentTask;
import bg.unisofia.fmi.grosd.ns.peristent_task.impl.NSJavaPersistentTaskFactory;
import bg.unisofia.fmi.grosd.ns.processor.NSTaskProcessor;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskUpload;
import bg.unisofia.fmi.grosd.ns.repository.NSRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepository;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.result.NullNSTaskResultImpl;
import bg.unisofia.fmi.grosd.ns.util.NSTaskLifecycleNotifier;
import bg.unisofia.fmi.grosd.ns.util.NodeServiceUtilities;


/**
 * Processor unit for uploading a new persitent task to the current node.
 *@author Shenol Yousouf
 */
public class NSJavaPersistentTaskUploadProcessor extends Thread implements NSTaskProcessor {
  
  private NodeServiceClientCallback client;
  private NSJavaPersistentTaskUpload task;
  private final String taskID;
  private final NSTaskRepositoryRecord taskRecord;
  
  private final Set<File> filesForGC = new HashSet<File>();
  
  
  public NSJavaPersistentTaskUploadProcessor (NSJavaPersistentTaskUpload task, NodeServiceClientCallback client) throws RemoteException {
    this.task = task;
    this.taskID = task.getTaskID();
    this.client = client;
    this.taskRecord = NSTaskRepository.registerTask(taskID, this);    
  }

  public void stopTask(boolean synchronous) throws NodeServiceException {
    throw new NodeServiceException("Stopping the uplooad of Node Service Java Persistent Task is not supported.");
  }

  public void run() {
    try {
      String keyIdentifier = task.getKeyIdentifier();
      NSRepositoryRecord record = NSTaskRepository.getServiceRecord(keyIdentifier);
      if (record != null) {
        // with the current implementation of Node Service classloader, the task binaries stay locked until the restart of Node Service;
        // that's why any update of the binaries will fail when the Node Service is running
        throw new NodeServiceException("There is already a persistent task for key " + keyIdentifier + " that is active on this node.\r\n"
                            + "The persistent task can be replaced only after removal of the current one and a restart of the Node Service.");        
      }
      
      notifyTaskStarted();
    } catch (Exception e) {
      // TODO: log fatal error - the only possible reason for the exception is inconsistent start task state
      e.printStackTrace();
      
      notifyTaskAborted(e);
      return;
    }
    
    // step 1: build persistent task object from remote task object so that it can be serialized
    // step 2: persist task so that the NodeService can recover it after restart
    // step 3: if all else is successful, register persistent task in Node Service Repository
    
    
    NSJavaPersistentTask persistentTask;
    try {
      // step 1/3
      persistentTask = NSJavaPersistentTaskFactory.buildNSPersistentTask(task, client, filesForGC);
      
      // step 2/3
      // if we get here, then the persistent task should have been initialized successfully
      persistTask(persistentTask);
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();
      
      // since the task upload has failed, all files for GC are registered to current task
      taskRecord.addFilesForGC(filesForGC.toArray(new File[] {}));
      
      notifyTaskAborted(e);
      return;
    }
    
    // step 3/3
    NSRepositoryRecord persistentTaskRecord;
    try {
      // no class loading is required here because this is done on a previous step
      persistentTaskRecord = NSTaskRepository.registerService(persistentTask.getKeyID(), persistentTask, false);
      persistentTaskRecord.addFilesForGC(filesForGC.toArray(new File[] {}));
    } catch (NSClassLoaderException e) {
      // exception is generated only if the classloading feature is enabled
    }
    
    notifyTaskFinished();
  }

  private void notifyTaskAborted(Exception ex)  {
    try {
      taskRecord.setState(ABORTED);
    } catch (NodeServiceException e) {
      // TODO: log  error
      e.printStackTrace();
    }
    
    try {
      task.changeActionState(ABORTED, "Task with ID " + taskID + " has been aborted. Reason: " + NodeServiceUtilities.formatExceptionMessage(ex));
      NSTaskLifecycleNotifier.notifyTaskAborted(task, "Task with ID " + taskID + "  aborted on start - see exception for cause.", new Exception[] {ex});
    } catch (Exception e) {
      // TODO: log  error
      e.printStackTrace();
    }
  }
  
  private void notifyTaskStarted() throws NodeServiceException {

    taskRecord.setState(STARTED);
    try {
     task.changeActionState(STARTED, "Task with ID " + taskID + " successfully started.");
    } catch (RemoteException e) {
      // TODO: log warning / error; no need to abort task for failing to change its state
      e.printStackTrace();
    }    
  }
  
  private void notifyTaskFinished() {
    
    try {
      taskRecord.setState(FINISHED);
    } catch (NodeServiceException e) {
      // TODO: log error - no need to abort the task at the end
      e.printStackTrace();
    }
    
    try {
      task.changeActionState(FINISHED, "Task with ID " + taskID + " finished successfully.");
      NSTaskLifecycleNotifier.notifyTaskFinished(task, new NullNSTaskResultImpl());
    } catch (RemoteException e) {
      // TODO: log  error
      e.printStackTrace();
    }
  }
  
  private void persistTask(NSJavaPersistentTask persistentTask) throws IOException {
    String keyIdentifier = persistentTask.getKeyID();
    File persistentTaskFile = NodeServiceUtilities.getFileForPersistentTask(keyIdentifier);
    ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(persistentTaskFile));
    try {
      out.writeObject(persistentTask);
    } finally {
      // make sure to close even iin case of exception
      out.close();
    }
  }

  public String getTaskID() {
    return taskID;
  }
}
