
package bg.unisofia.fmi.grosd.ns.processor.impl;

import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.ABORTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.FINISHED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STARTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STOPPED;

import java.io.File;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashSet;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.application.StoppableApplication;
import bg.unisofia.fmi.grosd.ns.exception.NSJavaTaskProcessorException;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.processor.NSTaskProcessor;
import bg.unisofia.fmi.grosd.ns.processor.visitor.NSMethodArgumentVisitor;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaTask;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.NSTaskArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.visitor.NSTaskArgumentVisitor;
import bg.unisofia.fmi.grosd.ns.remote.task.result.locator.NSFileResultLocator;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepository;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.result.NSTaskResultImpl;
import bg.unisofia.fmi.grosd.ns.util.NSClassChecker;
import bg.unisofia.fmi.grosd.ns.util.NSFileTranferUtilities;
import bg.unisofia.fmi.grosd.ns.util.NSTaskLifecycleNotifier;
import bg.unisofia.fmi.grosd.ns.util.NodeServiceUtilities;


/**
 * Processor unit for executing a Java task - invocation of a specified method with specified parameters
 * in a specified class.
 *@author Shenol Yousouf
 */
public class NSJavaTaskProcessor extends Thread implements NSTaskProcessor {
  
  private final NSJavaTask task;
  private final String taskID;
  private final NSTaskRepositoryRecord taskRecord;
  private TaskExecutionThread taskThread;
  
  private final NodeServiceClientCallback client;
  
  //private SynchonizedActionState taskState;
  //private boolean inProgress;
  //private boolean aborted;
  //private boolean finished;         // implies finished successfully
  
  private final File workFolder;
  private final File currentFolder;
  
  private final Set<File> filesForGC = new HashSet<File>();
  
  // this flag is raised only when the task execution should be stopped
  private boolean stopping = false;
  
  public NSJavaTaskProcessor(NSJavaTask task, NodeServiceClientCallback client) throws RemoteException {
    this.task = task;
    this.client = client;
    this.taskID = task.getTaskID();
    //this.taskState = new SynchonizedActionState(NSTaskActionState.WAITING);
    //this.inProgress = false;
    //this.aborted = false;
    //this.finished = false;
    this.workFolder = NodeServiceUtilities.getWorkFolderForTask(taskID);
    this.currentFolder = new File (System.getProperty("user.dir"));
    this.taskRecord = NSTaskRepository.registerTask(taskID, this);
  }
  
  /**
   * Initializa the required data for further processing of the Java task.
   * Transfer all required external files.
   * @throws RemoteException
   */
  private void initializeTask() throws RemoteException {    
    
    if (workFolder.exists() == false && workFolder.mkdirs() == false) {
      throw new NSJavaTaskProcessorException("Cannot create task's work folder - " + workFolder.getAbsolutePath());
    }
    
    // take care for the transfer of necessary files, first
    
    // binary files are not marked for GC at this point because they are copied into the work folder
    // which is deleted in all cases
    NSFileTranferUtilities.transferClientFiles(null, client, workFolder, task.getBinaryFiles(), true);
    NSFileTranferUtilities.transferClientFiles(filesForGC, client, currentFolder, task.getAdditionalFiles(), false);
  }
  
  /**
   * Process the task metadata and execute the task.
   * @throws RemoteException
   */
  private void performTask() throws RemoteException {
    
    // replacing the context classloader may solve issues when unwrapping argument values
    Thread.currentThread().setContextClassLoader(NodeServiceInternal.getClassLoader());
    
    String classname = task.getTargetClass().getClassName();
    if (classname == null || classname.length() == 0) {
      throw new NSJavaTaskProcessorException("The provided class name for executing the task " + taskID + "  is not defined.");
    }
    
    
    // load class for task
    Class c = null;
    try {
      c = Class.forName(classname, true, NodeServiceInternal.getClassLoader());
    } catch (Exception e) {
      throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error loading target class for name " + classname, e);
    }
    
    
    // construct task method for invocation
    String methodName = task.getMethod().getMethodName();    
    Class[] parameterTypes = null;
    
    String[] parameterClassnames = task.getMethod().getParameterClassnames();
    if (parameterClassnames != null && parameterClassnames.length > 0) {
      parameterTypes = new Class[parameterClassnames.length];
      byte index = 0;
      Class parameterClass = null;
      for(String parameterClassname : parameterClassnames) {
        try {
          parameterClass = NSClassChecker.parseClassname(parameterClassname);
        } catch (Exception e) {
          throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error loading parameter class for name " + parameterClassname, e);
        }
        parameterTypes[index++] = parameterClass;  
      }
    }
    
    NSTaskArgument[] arguments = task.getMethod().getArgs();
    NSMethodArgumentVisitor visitorImpl = new NSMethodArgumentVisitor(client, filesForGC, currentFolder, (arguments != null) ? arguments.length : 0, (parameterTypes == null));
    
    if (arguments != null) {
      NSTaskArgumentVisitor visitor = (NSTaskArgumentVisitor) UnicastRemoteObject.exportObject(visitorImpl, 0);
      for(NSTaskArgument argument : arguments) {
        argument.accept(visitor);
      }
    }
    
    if (parameterTypes == null) {   // read types from the argument classes
     parameterTypes = visitorImpl.getClasses();
    }
    
    Method method = null;
    try {
      method = c.getDeclaredMethod(methodName, parameterTypes);
    } catch (SecurityException e) {
      throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error getting method from class + " + classname, e);
    } catch (NoSuchMethodException e) {
      Method[] matches = NSClassChecker.searchForMethod(c, methodName, parameterTypes);
      if (matches.length != 1) {
        throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Cannot find exact match for method with name " + methodName + " in class " + classname);
      }
      method = matches[0];
    }
    
    
    // prepare for method invocation
    Object instance = null;
    if (Modifier.isStatic(method.getModifiers()) == false) { // this is not a static method and an object has to be instantiated
                                                                                                 // on which the method will be invoked
      NSTaskArgument[] initArgs = task.getTargetClass().getInitArgs();
      if (initArgs == null || initArgs.length == 0) {                   // no arguments ---> use default constructor
        try {
          instance = c.newInstance();
        } catch (Exception e) {
          throw new NSJavaTaskProcessorException("Error instantiating object for method invocation.", e);
        }
      } else {                                                                             // get the required constructor
        
        Class[] consParameterTypes = null;
        String[] consParameterTypeNames = task.getTargetClass().getParameterClassnames();
        byte index = 0;
        Class consParameterClass = null;
        if (consParameterTypeNames != null && consParameterTypeNames.length > 0) {
          consParameterTypes = new Class[consParameterTypeNames.length];
          for (String consParameterTypesName : consParameterTypeNames) {
            try {
              consParameterClass = NSClassChecker.parseClassname(consParameterTypesName);
            } catch (Exception e) {
              throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error loading constructor parameter class for name " + consParameterTypesName, e);
            }
            consParameterTypes[index++] = consParameterClass;
          }
        }
        
        NSMethodArgumentVisitor consVisitorImpl = new NSMethodArgumentVisitor(client, filesForGC, currentFolder, initArgs.length, (consParameterTypes == null));
        NSTaskArgumentVisitor consVisitor = (NSTaskArgumentVisitor) UnicastRemoteObject.exportObject(consVisitorImpl, 0);
         for(NSTaskArgument initArg:initArgs) {
           initArg.accept(consVisitor);
         }
         
         if (consParameterTypes == null) {      // read it from argument types
           consParameterTypes = consVisitorImpl.getClasses();
         }
         
         Constructor cons = null;
         try {
          cons = c.getConstructor(consParameterTypes);
        } catch (SecurityException e) {
          throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error instantiating object for method invocation", e);
        } catch (NoSuchMethodException e) {
          Constructor[] matches = NSClassChecker.searchForConstructor(c, consParameterTypes);
          if ( matches.length != 1) {
            throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Cannot find exact match for constructor for class " + classname + " with given initial paramters.");
          }
          cons = matches[0];
        }
        
        try {
          instance = cons.newInstance(consVisitorImpl.getArgValues());
        } catch (Exception e) {
          throw new NSJavaTaskProcessorException("TaskID " + taskID + ": Error instantiating object for method invocation woth provided initial arguments.", e);
        }
      }
    }
    
    
    // time to invoke task
    taskThread = new TaskExecutionThread(method, instance, visitorImpl.getArgValues());
    taskThread.start();
  }
  
  

  public void run() {
    try {
      initializeTask();
      performTask();
    } catch (Exception e) {
      // TODO: log error
      e.printStackTrace();
      notifyTaskAborted(e);
    } finally {
      registerFilesForGC(workFolder);
      registerFilesForGC(filesForGC.toArray(new File[] {}));
    }
  }
  
  public synchronized void stopTask(boolean synchronous) {
    if (taskRecord.isActionStateFinal()) {
      // task is already stopped
      return;
    }
    
    if (taskRecord.getState() == STARTED) {
      // task execution is already in progress and task thread is delegated to stop it
      taskThread.stopTask();
      notifyTaskStopped();
    } else {
      // execution has not bugun yet;
      // the next lock on the current processor will be acquired by the task thread when calling notifyTaskStarted();
      // for this purpose, the stopping flag is raised to signal that the execution should be terminated
      stopping = true;
      
      if (synchronous) {
        // this piece of code intends to make the call of stopTask synchronous with the actual task termination,
        // i.e. the method will return only after the task is really stopped
        while(taskRecord.isActionStateFinal() == false) {
          try {
            wait();
          } catch (InterruptedException e) {
            // TODO: log  error
            e.printStackTrace();
          }
        }
      }
    }
  }
  
synchronized void notifyTaskStarted() throws NodeServiceException {
    
      // this is the only method for notifying that the task action state has changed to a non-final one;
      // for this purpose, it is necessary to check if there is a stop request only here.
      if (stopping) {
        taskThread.stopTask();
        notifyTaskStopped();
        return;
      }
      
      // inProgress = true;
      //taskState.setActionState(NSTaskActionState.STARTED);
      taskRecord.setState(STARTED);
      try {
       task.changeActionState(STARTED, "Task with ID " + taskID + " successfully started.");
      } catch (RemoteException e) {
        // TODO: log warning / error; no need to abort task for failing to change its state
        e.printStackTrace();
      }    
  }
  
  synchronized void notifyTaskStopped()  {
    
    try {
      // if task is already in final state, do not propagate the exception to upper layers -
      // task is in stopped state, anyway
      taskRecord.setState(STOPPED);
      // notify any synchronous listeners
      notifyAll();
      task.changeActionState(STOPPED, "Task with ID " + taskID + " stopped on request.");
    } catch (RemoteException e) {
      // TODO: log  error
      e.printStackTrace();
    }
  }
  
  synchronized void notifyTaskAborted(Exception ex) {
    //aborted = true;
    //inProgress = false;
    //taskState.setActionState(NSTaskActionState.ABORTED);

    try {
      taskRecord.setState(ABORTED);
    } catch (NodeServiceException e) {
      // TODO: log  error
      e.printStackTrace();
    }
    
    try {
      task.changeActionState(ABORTED, "Task with ID " + taskID + " has been aborted. Reason: " + NodeServiceUtilities.formatExceptionMessage(ex));
      NSTaskLifecycleNotifier.notifyTaskAborted(task, "Task with ID " + taskID + "  aborted on start - see exception for cause.", new Exception[] {ex});
    } catch (Exception e) {
      // TODO: log  error
      e.printStackTrace();
    }
    
    
  }
  
  synchronized void notifyTaskFinished(Serializable result) throws NodeServiceException {
    //inProgress = false;
    NSFileResultLocator fileResultLocator = null;
    File[] fileResults = null;
    NSTaskResultImpl taskResult = null;
    StringBuilder resultComment = new StringBuilder();
    
    try {
      fileResultLocator = task.getFileResultLocator();
    } catch (RemoteException e) {
      // TODO: log error; an attempt will be made to transfer at least the result
      e.printStackTrace();
      resultComment.append("Error in retrieving locator for file results\nCaused by: " + NodeServiceUtilities.formatExceptionMessage(e));
    }
    
    if (fileResultLocator != null) {
      try {
        StringBuilder searchComments = new StringBuilder();
        fileResults = NodeServiceUtilities.getOutputFiles(currentFolder, fileResultLocator, searchComments);
        resultComment.append(searchComments.toString());
      } catch (Exception e) {
        resultComment.append("Error in getting file results.\nCaused by: " + NodeServiceUtilities.formatExceptionMessage(e));
      }
    }
    
   
    if (fileResults == null) {
      if (fileResultLocator != null) {
        resultComment.append("No file results were found ! Possible error in task execution !");
      }
      taskResult = new NSTaskResultImpl(result, currentFolder, new File[] {}, client, resultComment.toString());
    } else {
      registerFilesForGC(fileResults);
      taskResult = new NSTaskResultImpl(result, currentFolder, fileResults, client, resultComment.toString());
    }
    
    //this.finished = true;
    //taskState.setActionState(NSTaskActionState.FINISHED);
    
    taskRecord.setState(FINISHED);
    try {
      task.changeActionState(FINISHED, "Task with ID " + taskID + " finished successfully.");
      NSTaskLifecycleNotifier.notifyTaskFinished(task, taskResult);
    } catch (RemoteException e) {
      // TODO: log  error
      e.printStackTrace();
    }
  }
  
  void registerFilesForGC(File...files) {
    taskRecord.addFilesForGC(files);
  }
 
  /**
   * A helper thread for carrying out the actual method invocation, after all the initial metadata
   * has been processed and the method for invocation was constucted.
   * The execution in thread also creates a unit for control of the execution lidecycle itself. 
   * @author Shenol Yousouf
   */
  class TaskExecutionThread extends Thread {
    private Method method;
    Object instance;
    private Object[] arguments;
    
    public TaskExecutionThread(Method method, Object instance, Object... arguments) {
      this.method = method;
      this.instance = instance;
      this.arguments = arguments;
    }
    
    public void run() {
      try {
        // this notification is not entirely correct, because we have not invoked the method yet;
        // however, this will prevent confusing situations when client is notified that the task is finished or aborted
        // even before being notified that task was started
        notifyTaskStarted();
      
        Serializable result = null;
        try {
          result = (Serializable) method.invoke(instance, arguments);
        } catch (Exception e) {
          notifyTaskAborted(e);
          return;
        }
        if (result != null && result instanceof File) {
          registerFilesForGC((File)result);
        }
        notifyTaskFinished(result);
      } catch (NodeServiceException e) {
        // TODO: log  error
        notifyTaskAborted(e);
      }
    }
    
    @SuppressWarnings("deprecation")
    public  void stopTask() {
      if (instance instanceof StoppableApplication) {
       ((StoppableApplication) instance).stopExecution();
      }
      // though not elegant, it will ensure stopping of this thread
      this.stop();
    }
    
  }

  public String getTaskID() {
    return taskID;
  }

}
