
package bg.unisofia.fmi.grosd.ns.repository;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import bg.unisofia.fmi.grosd.ns.exception.NSClassLoaderException;
import bg.unisofia.fmi.grosd.ns.peristent_task.NSJavaPersistentTask;
import bg.unisofia.fmi.grosd.ns.processor.NSTaskProcessor;


/**
 * Acts as a registry for all active tasks and persistent services on node.
 * Java tasks are registered together with their execution processors and yask identificators
 * while services are registered with their key identifiers only for access since they are not
 * always employed by processor units. 
 *@author Shenol Yousouf
 */
public abstract class NSTaskRepository {

  private static Map<String, NSTaskRepositoryRecord> tasksMap;
  private static Map<String, NSRepositoryRecord<NSJavaPersistentTask>> servicesMap;
  
  public static void init() {
    tasksMap = new HashMap<String, NSTaskRepositoryRecord>();
    servicesMap = new HashMap<String, NSRepositoryRecord<NSJavaPersistentTask>>();
  }
  
  
  // tasks
  
  public static synchronized NSTaskRepositoryRecord registerTask(String taskID, NSTaskProcessor processor) {
    // TODO: check if this taskID already exists in registry;
    // with the current API, it is highly unprobable to have two tasks with the same ID, because they must be created exactly at the same millisecond,
    // but still this check should not be API-dependent  
    NSTaskRepositoryRecord record = new NSTaskRepositoryRecord(processor);
    tasksMap.put(taskID, record);
    return record;
  }
  
  public static synchronized NSTaskRepositoryRecord getTaskRecord(String taskID) {
    return tasksMap.get(taskID);
  }
  
  public static synchronized void removeTaskRecord(String taskID) {
    tasksMap.remove(taskID);
  }
  
  public static synchronized ArrayList<NSTaskProcessor> getAllRegisteredTaskProcessors() {
    
    ArrayList<NSTaskProcessor> result = new ArrayList<NSTaskProcessor>(tasksMap.size());
    Collection<NSTaskRepositoryRecord> records = tasksMap.values();
    for (NSTaskRepositoryRecord record : records) {
      result.add(record.getProcessor());
    }    
    return result;  
  }    
  
  
  
  // services
  
  public static synchronized NSRepositoryRecord<NSJavaPersistentTask> registerService(String keyID, NSJavaPersistentTask service, boolean withClassloading) throws NSClassLoaderException {
    // TODO: check if this taskID already exists in registry
    NSRepositoryRecord<NSJavaPersistentTask> record = new NSRepositoryRecord<NSJavaPersistentTask>(service);
    if (withClassloading) {
      service.loadServiceClasses();
    }
    servicesMap.put(keyID, record);
    return record;
  }
  
  public static synchronized NSRepositoryRecord<NSJavaPersistentTask> getServiceRecord(String keyID) {
    return servicesMap.get(keyID);
  }
  
  public static synchronized void removeServiceRecord(String keyID) {
    servicesMap.remove(keyID);
  }
  
  public static synchronized ArrayList<NSJavaPersistentTask> getAllRegisteredPersistentTasks() {
    
    ArrayList<NSJavaPersistentTask> result = new ArrayList<NSJavaPersistentTask>(servicesMap.size());
    Collection<NSRepositoryRecord<NSJavaPersistentTask>> records = servicesMap.values();
    for (NSRepositoryRecord<NSJavaPersistentTask> record : records) {
      result.add(record.getProcessor());
    }
    return result;
  }
}
