
package bg.unisofia.fmi.grosd.ns.result;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;

import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.result.NSTaskResult;
import bg.unisofia.fmi.grosd.ns.util.NSFileTranferUtilities;


/**
 * @see NSTaskResult
 *@author Shenol Yousouf
 */
public class NSTaskResultImpl implements NSTaskResult {
  
  private Serializable result;
  private File[] files;
  private NodeServiceClientCallback client;
  private File root;
  private String comment;
  
  public NSTaskResultImpl(Serializable result, File root, File[] files, NodeServiceClientCallback client, String comment) {
    this.result = result;
    this.files = files;
    this.client = client;
    this.root = root;
    this.comment = comment;
  }

  public Serializable getResult() throws RemoteException {
    return result;
  }

  public byte getResultAsByte() throws RemoteException {
    String str = getResultAsString();
    try {
      return Byte.parseByte(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("byte"), e);
    }
  }

  public char getResultAsChar() throws RemoteException {
    if (result instanceof Character) {
      return ((Character) result).charValue();
    } else {
      throw new RemoteException(exceptionString("char"));
    }
  }

  public double getResultAsDouble() throws RemoteException {
    String str = getResultAsString();
    try {
      return Double.parseDouble(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("double"), e);
    }
  }

  public URL[] getResultAsFileURLs() throws RemoteException {
    ArrayList<URL> urls = new ArrayList<URL>();
    
    if (result != null && result instanceof File) {
      File file = (File) result;
        if (file.exists()) {
        try {
          copyContents(urls, file);
        } catch (IOException e) {
          throw new RemoteException("Error getting file result.", e);
        }
      } else {
        // there are no file contents for transfer
        // TODO: log warning; maybe the file intentionally exists only as a file descriptor
      }
    }
    
    if (files != null) {
      for (File file:files) {
        if (file.exists()) {
          try {
            copyContents(urls, file);
          } catch (IOException e) {
            throw new RemoteException("Error getting file result.", e);
          }
        } else  {
          // there are no file contents for transfer
          // TODO: log warning; maybe the file intentionally exists only as a file descriptor
        }
      }
    }
    
    return urls.toArray(new URL[] {});
  }

  public float getResultAsFloat() throws RemoteException {
    String str = getResultAsString();
    try {
      return Float.parseFloat(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("float"), e);
    }
  }

  public int getResultAsInt() throws RemoteException {
    String str = getResultAsString();
    try {
      return Integer.parseInt(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("integer"), e);
    }
  }

  public long getResultAsLong() throws RemoteException {
    String str = getResultAsString();
    try {
      return Long.parseLong(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("long"), e);
    }
  }

  public short getResultAsShort() throws RemoteException {
    String str = getResultAsString();
    try {
      return Short.parseShort(str);
    } catch (NumberFormatException e) {
      throw new RemoteException(exceptionString("short"), e);
    }
  }

  public boolean getReultAsBoolean() throws RemoteException {
    String str = getResultAsString();
    return Boolean.valueOf(str);                    // TODO: do we need a more strict implementation ?
  }
  
  private String getResultAsString() throws RemoteException {
    if (result == null) {
      throw new RemoteException("Result is null.");
    }
    return result.toString();
  }
  
  private String exceptionString(String type) {
    return "Result " + result + " cannot be parsed  to " + type + " - result type is " + result.getClass().getName() + ".";
  }
  
  private void copyContents(ArrayList<URL> urls, File file) throws IOException {
    if (file.isFile()) {
      URL url = NSFileTranferUtilities.copyToRemoteLocaltion(client, file, contructDestinationFileName(file));
      urls.add(url);
    } else if (file.isDirectory()) {
      File[] files = file.listFiles();
      for (int i = 0; i < files.length; i++) {
        copyContents(urls, files[i]);
      }
    }
  }
  
  /**
   * @param file file for processing
   * @return the relative path of file against the root folder if file is located in a folder hiearcrhy
   * beneath the root folder, stripping any starting file separator chars; otherwise, only the file name is returned
   */
  private String contructDestinationFileName (File file) {
    String rootFolderName = root.getAbsolutePath();
    String filePath = file.getAbsolutePath();
    if (filePath.startsWith(rootFolderName)) {
      filePath = filePath.substring(rootFolderName.length());
      if (filePath.startsWith(File.separator)) {
        filePath = filePath.substring(1);
      }
    } else {
      filePath = file.getName();
    }
    return filePath;
  }

  public String getComment() throws RemoteException {
    return comment;
  }

}
