
package bg.unisofia.fmi.grosd.ns.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;


/**
 * Class for verifying the parameters of customer classes and for resolving the target methods
 * and initialization constructors by the provided lists of input parameters classnames.
 *@author Shenol Yousouf
 */
public abstract class NSClassChecker {
  
  private final static Map<Class, Set<Class>> autoboxedTypesIndex = new HashMap<Class, Set<Class>>();
  private final static Map<Class, Class> primitiveTypeMappings = new HashMap<Class, Class>();
  private final static Map<String, Class> primitiveTypeNameMappings = new HashMap<String, Class>();
  
  static {
    
    final Set<Class> booleanSet = new HashSet<Class>();
    booleanSet.add(boolean.class);
    
    final Set<Class> byteSet = new HashSet<Class>();
    byteSet.add(byte.class);
    
    final Set<Class> charSet = new HashSet<Class>();
    charSet.add(char.class);

    final Set<Class> shortSet = new HashSet<Class>(byteSet);
    shortSet.add(short.class);

    final Set<Class> intSet = new HashSet<Class>(shortSet);
    intSet.add(char.class);
    intSet.add(int.class);

    final Set<Class> longSet = new HashSet<Class>(intSet);
    longSet.add(long.class);

    final Set<Class> floatSet = new HashSet<Class>(longSet);
    floatSet.add(float.class);
    
    final Set<Class> doubleSet = new HashSet<Class>(floatSet);
    doubleSet.add(double.class);
    
    autoboxedTypesIndex.put(byte.class, byteSet);
    autoboxedTypesIndex.put(Byte.class, byteSet);
    
    autoboxedTypesIndex.put(char.class, charSet);
    autoboxedTypesIndex.put(Character.class, charSet);
    
    autoboxedTypesIndex.put(short.class, shortSet);
    autoboxedTypesIndex.put(Short.class, shortSet);
    
    autoboxedTypesIndex.put(int.class, intSet);
    autoboxedTypesIndex.put(Integer.class, intSet);
    
    autoboxedTypesIndex.put(long.class, longSet);
    autoboxedTypesIndex.put(Long.class, longSet);
    
    autoboxedTypesIndex.put(float.class, floatSet);
    autoboxedTypesIndex.put(Float.class, floatSet);
    
    autoboxedTypesIndex.put(double.class, doubleSet);
    autoboxedTypesIndex.put(Double.class, doubleSet);
    
    autoboxedTypesIndex.put(boolean.class, booleanSet);
    autoboxedTypesIndex.put(Boolean.class, booleanSet);
    
    primitiveTypeMappings.put(Byte.class, byte.class);
    primitiveTypeMappings.put(Short.class, short.class);
    primitiveTypeMappings.put(Character.class, char.class);
    primitiveTypeMappings.put(Integer.class, int.class);
    primitiveTypeMappings.put(Long.class, long.class);
    primitiveTypeMappings.put(Float.class, float.class);
    primitiveTypeMappings.put(Double.class, double.class);
    primitiveTypeMappings.put(Boolean.class, boolean.class);
    
    primitiveTypeNameMappings.put(byte.class.getName(), byte.class);
    primitiveTypeNameMappings.put(short.class.getName(), short.class);
    primitiveTypeNameMappings.put(char.class.getName(), char.class);
    primitiveTypeNameMappings.put(int.class.getName(), int.class);
    primitiveTypeNameMappings.put(long.class.getName(), long.class);
    primitiveTypeNameMappings.put(float.class.getName(), float.class);
    primitiveTypeNameMappings.put(double.class.getName(), double.class);
    primitiveTypeNameMappings.put(boolean.class.getName(), boolean.class);
    
  }

  /**
   * 
   * @param c class inside which the constructor is looked for
   * @param parameterTypes list of input parameters classes for identifying the proper constructor
   * @return resolved constructor
   */
  public static Constructor[] searchForConstructor(Class c, Class[] parameterTypes) {
    ArrayList<Constructor> matches = new ArrayList<Constructor>();
    Constructor[] potentialMatches = c.getConstructors();
    for(Constructor cons:potentialMatches) {
      if (cons.getParameterTypes().length == parameterTypes.length) {
        boolean matchFound =true;
        Class[] types = cons.getParameterTypes();
        for (int i = 0; i < types.length; i++) {
          if (areAssignable(types[i], parameterTypes[i]) == false) {
            matchFound = false;
            break;
          }
        }
        if (matchFound) {
          matches.add(cons);
        }
      }
    }
    
    if (matches.size() > 1) {
      matches = narrowDownSearchForConstructor(matches, parameterTypes);
    }
    
    return matches.toArray(new Constructor[] {});
  }

  /**
   * 
   * @param c class inside which the method is looked for
   * @param methodName method name
   * @param parameterTypes list of input parameters classes for identifying the proper method
   * @return resolved method
   */
  public static Method[] searchForMethod(Class c, String methodName, Class[] parameterTypes) {
    ArrayList<Method> matches = new ArrayList<Method>();
    Method[] potentialMatches = c.getDeclaredMethods();
    for(Method method:potentialMatches) {
      if (method.getName().equals(methodName)
                          && method.getParameterTypes().length == parameterTypes.length) {
        boolean matchFound =true;
        Class[] types = method.getParameterTypes();
        for (int i = 0; i < types.length; i++) {
          if (areAssignable(types[i], parameterTypes[i]) == false) {
            matchFound = false;
            break;
          }
        }
        if (matchFound) {
          matches.add(method);
        }
      }
    }

    if (matches.size() > 1) {
      matches = narrowDownSearchForMethod(matches, parameterTypes);
    }
    
    return matches.toArray(new Method[] {});
  }
  
  /**
   * Makes an extensive check if an argument of type <code>lclass</code> is assignable from
   * an argument of type <code>rclass</code> by taking into account the autoboxing feature in JDK 5
   * and class hierarchies.
   * @param c1 type of the left-side argument in assignment
   * @param c2 type of the right-side argument in assignment
   * @return <code>true</code> if argument of type<code>rclass</code> can be assigned to argument of type <code>lclass</code>
   */
  public static boolean areAssignable(Class lclass, Class rclass) {
    
    if (lclass == null || rclass == null) {
      return false;
    }
    
    if (lclass.isAssignableFrom(rclass)) {
      return true;
    }
    
    if (lclass.isPrimitive() == false && rclass.isPrimitive() == false) {
      return false;
    }
    
    if (rclass.isPrimitive() == false) {                    // then lclass must be primitive...
      rclass = primitiveTypeMappings.get(rclass);
    }
    
    Set typeIndex = autoboxedTypesIndex.get(lclass);
    
    if (rclass == null || typeIndex == null) {
      return false;
    }
    
    return typeIndex.contains(rclass);
  }
  
  private static ArrayList<Method> narrowDownSearchForMethod(ArrayList<Method> potentialMatches, Class[] parameterTypes) {
    ArrayList<Method> result = new ArrayList<Method>();
    byte currentBestCount = 100;
    
    for(Method potentialMatch : potentialMatches) {
      byte diffCount = getDifferenceCount(potentialMatch.getParameterTypes(), parameterTypes);
      
      if (diffCount == 0) {     // exact match !
        result.clear();
        result.add(potentialMatch);
        break;
      }
      
      if (diffCount < currentBestCount) {
        result.clear();
        result.add(potentialMatch);
        currentBestCount = diffCount;
      }
      
      if (diffCount == currentBestCount) {
        result.add(potentialMatch);
      }
    }
    
    return result;
  }
  
  private static ArrayList<Constructor> narrowDownSearchForConstructor(ArrayList<Constructor> potentialMatches, Class[] parameterTypes) {
    ArrayList<Constructor> result = new ArrayList<Constructor>();
    byte currentBestCount = 100;
    
    for(Constructor potentialMatch : potentialMatches) {
      byte diffCount = getDifferenceCount(potentialMatch.getParameterTypes(), parameterTypes);
      
      if (diffCount == 0) {     // exact match !
        result.clear();
        result.add(potentialMatch);
        break;
      }
      
      if (diffCount < currentBestCount) {
        result.clear();
        result.add(potentialMatch);
        currentBestCount = diffCount;
      }
      
      if (diffCount == currentBestCount) {
        result.add(potentialMatch);
      }
    }
    
    return result;
  }
  
  private static byte getDifferenceCount (Class[] clarr1, Class[] clarr2) {
    byte diffCount = 0;
    for(int index = 0; index < clarr1.length; index++) {    // it is supposed that clarr1 and clarr2 have the same length
      if (clarr1[index].equals(clarr2[index]) == false) {
        diffCount++;
      }
    }
    return diffCount;
  }
  
  /**
   * 
   * @param classname name of the class inside which the constructor is looked for
   * @param parameterClassnames list of input parameters class names for identifying the proper constructor
   * @return resolved constructor
   */
  public static Constructor parseConsSignature(String classname, String[] parameterClassnames) throws ClassNotFoundException, NodeServiceException {

    Constructor cons = null;
    
    if (parameterClassnames!= null && parameterClassnames.length > 0) {
      // load parameter classes
      Class[] initializationClasses = new Class[parameterClassnames.length];
      for (int index = 0; index < parameterClassnames.length; index++) {
        initializationClasses[index] = parseClassname(parameterClassnames[index]);          
      }
    
      cons = parseConsSignature(classname, initializationClasses);
    }
    
    return cons;
  }
  
  /**
   * 
   * @param classname name of the class inside which the constructor is looked for
   * @param initializationClasses list of input parameters classes for identifying the proper constructor
   * @return resolved constructor
   */
  public static Constructor parseConsSignature(String classname, Class[] initializationClasses) throws ClassNotFoundException, NodeServiceException {
    
    Constructor cons = null;

    Class c = parseClassname(classname);
    
    try {
      cons = c.getConstructor(initializationClasses);
    } catch (SecurityException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NodeServiceException("Cannot get constructor for class " + c.getName(), e);
    } catch (NoSuchMethodException e) {
      Constructor[] matches = searchForConstructor(c, initializationClasses);
      if (matches.length != 1) {
        throw new NodeServiceException("Cannot find exact match for constructor for class " + c.getName() + " with given initial paramters.");
      }
      cons = matches[0];
    }
    
    return cons;
  }
  
  /**
   * 
   * @param classname name of the class inside which the method is looked for
   * @param methodName method name
   * @param parameterClassnames list of input parameters class names for identifying the proper method
   * @return resolved method
   */
  public static Method parseMethodSignature(String classname, String methodName, String[] parameterClassnames) throws ClassNotFoundException, NodeServiceException {
    
    Method method = null;
    
    Class[] parametersClasses = (parameterClassnames != null) ? new Class[parameterClassnames.length] : new Class[] {};
        
    for (int index = 0; index < parametersClasses.length; index++) {
      parametersClasses[index] = parseClassname(parameterClassnames[index]);
    }
    
   method = parseMethodSignature(classname, methodName, parametersClasses);
    
    return method;
  }
  
  /**
   * 
   * @param classname name of the class inside which the method is looked for
   * @param methodName method name
   * @param parametersClasses list of input parameters classes for identifying the proper method
   * @return resolved method
   */
  public static Method parseMethodSignature(String classname, String methodName, Class[] parametersClasses) throws ClassNotFoundException, NodeServiceException {
    Method method = null;

    Class c = parseClassname(classname);
    try {
      method = c.getDeclaredMethod(methodName, parametersClasses);
    } catch (SecurityException e) {
      // TODO: log error
      e.printStackTrace();
      throw new NodeServiceException("Cannot get method with name " + methodName + " for class " + classname, e);
    } catch (NoSuchMethodException e) {
      Method[] matches = searchForMethod(c, methodName, parametersClasses);
      if (matches.length != 1) {
        throw new NodeServiceException("Cannot find exact match for method with name " +  methodName + " for class " + c.getName() + " with given initial paramters.");
      }
      method = matches[0];
    }
    
    return method;
  }
  
  /**
   * 
   * @param classname input class name
   * @return get <code>Class</code> for provided name;
   * NOTE that this current method is able to get the class of a primitive argument as well
   * @throws ClassNotFoundException
   */
  public static Class parseClassname(String classname) throws ClassNotFoundException {
    // check if this is primitive type
    Class c = primitiveTypeNameMappings.get(classname);
    // this is an object type, load it in the usual way
    if (c == null) {
      c = Class.forName(classname, true, NodeServiceInternal.getClassLoader());
    }
    return c;
  }

}
