
package bg.unisofia.fmi.grosd.ns.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.exception.NSClassLoaderException;
import bg.unisofia.fmi.grosd.ns.exception.NSFileTransferException;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.remote.NodeServiceClientCallback;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.FileArgument;
import bg.unisofia.fmi.grosd.ns.remote.util.NSOutputStreamWrapper;
import bg.unisofia.fmi.grosd.ns.remote.util.impl.NSOutputStreamWrapperImpl;


/**
 * Utility class for carrying out required file transfers with the customers hosts.
 *@author Shenol Yousouf
 */
public abstract class NSFileTranferUtilities {
  
  final static int BUFFER = 4096;
  public final static String JAR_EXTENTION = ".jar";
  public final static String CLASS_EXTENTION = ".class";
  
  /**
   * 
   * @param fileList list of files generated after transfer
   * @param client callback connection with the API
   * @param destinationFolder Node Service folder into which the files are transferred
   * @param fileArgs list client <code>FileArgument</code> instances
   * @param withClassloading if <code>true</code>, the transferred files are treated as binary files and are supplied to the <code>NodeServiceClassLoader</code>
   * @throws RemoteException
   */
  public static void transferClientFiles (Set<File> fileList, NodeServiceClientCallback client, File destinationFolder, FileArgument[] fileArgs, boolean withClassloading) throws RemoteException {
    
    if (fileArgs != null) {
      for(FileArgument fileArg:fileArgs) {
        File destination = constructDestinationFile(destinationFolder, fileArg);
        
        URL sourceURL = fileArg.getSourceFileURL();
        boolean copied = false;
        
        if (!fileArg.isLocal()) {    // direct access
          try {
            copyFileFromURL(sourceURL, destination);
            copied = true;                  // file is copied if we get here without Exception
          } catch (Exception e) {
            // do nothing, second attempt will be made
          }
        }
        
        // either the file is not directly accessible, or the first attempt for copy has failed
        if (copied == false) {
          try {
            copyFileFromClient(client, sourceURL, destination);
            copied = true;
          } catch (IOException e) {
            throw new NSFileTransferException("Error occurred while getting file from URL " + sourceURL, e);
          }
        }
        
        // if we get here, then file is already copied
        if (fileList != null) {
          fileList.add(destination);
        }
        
        if (withClassloading) {
          try {
            if (destination.getName().toLowerCase().endsWith(JAR_EXTENTION)) {
              NodeServiceInternal.getClassLoader().addURL(destination.toURI().toURL());
            } else if (destination.getName().toLowerCase().endsWith(CLASS_EXTENTION)) {
              NodeServiceInternal.getClassLoader().addURL(destinationFolder.toURI().toURL());
            }
          } catch (MalformedURLException e) {
            throw new NSClassLoaderException("Destination file cannot be added as URL to binary files for class loading.", e);
          }
        }
      }
    }
  }

  /**
   * Copy a single file from Node Service client.
   * @param client callback connection with the API
   * @param clientFile client file's URL
   * @param destination destination into which the file is copied
   * @throws IOException
   */
  public static void copyFileFromClient(NodeServiceClientCallback client, URL clientFile, File destination) throws IOException {
    
    NSOutputStreamWrapperImpl wrapperImpl = new NSOutputStreamWrapperImpl(destination);
    NSOutputStreamWrapper wrapper = (NSOutputStreamWrapper) UnicastRemoteObject.exportObject(wrapperImpl, 0);
    
    client.copyFile(wrapper, clientFile);
  }
  
  /**
   * Copy file directly from the provided URL.
   * @param clientFile client file's URL
   * @param destination destination into which the file is copied
   * @throws IOException
   * @throws URISyntaxException
   */
  public static void copyFileFromURL(URL clientFile, File destination) throws IOException, URISyntaxException {
    BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(destination), BUFFER);
    BufferedInputStream in = new BufferedInputStream(new FileInputStream(new File(clientFile.toURI())), BUFFER);
    
    byte[] b = new byte[BUFFER];
    int length = 0;
    try {
      while((length = in.read(b)) != -1) {
        out.write(b, 0, length);
      }
    } finally {
      out.close();
      
      try {
        in.close();
      } catch (IOException e) {
        // TODO: trace warning or something; this error is not fatal for the current task
        e.printStackTrace();
      }
    }
  }
  
  /**
   * Copy file from Node Service host to another remote location,
   * @param client callback connection with the API
   * @param source original file on Node Service host
   * @param fileName file name under which the file is persisted on client's host
   * @return URL to the copied file which is local for the destination host
   * @throws IOException
   */
  public static URL copyToRemoteLocaltion(NodeServiceClientCallback client, File source, String fileName) throws IOException {
    BufferedInputStream in = new BufferedInputStream(new FileInputStream(source), BUFFER);
    NSOutputStreamWrapper out = client.getOutputStream(fileName);
    byte[] b = new byte[BUFFER];
    int length = 0;
    while((length = in.read(b)) != -1) {
      out.write(b, length);
    }
    in.close();
    out.close();
    return out.getLocalOutputURL();
  }
  
  /**
   * 
   * @param folder folder inside which the file is created; if <code>null</code>, the original folder hierarchy on the host from which the file is coming, is recreated on the destiantion host
   * @param fileArg a <code>FileArgument</code> instance, which wraps the original client file
   * @return file reference on Node Service host
   * @throws RemoteException
   */
  public static File constructDestinationFile(File folder, FileArgument fileArg) throws RemoteException {
    String destination = fileArg.getRelativeDest();
    if (destination == null || destination.length() == 0) {
      destination = fileArg.getSourceFileURL().getPath();
    }
    File result = new File(folder, destination);
    File parentFile = result.getParentFile();
    if (parentFile != null && parentFile.exists() == false && (parentFile.mkdirs() == false)) {
      throw new NSFileTransferException("Error creating destination folder - " + parentFile.getAbsolutePath());
    }
    return result;
  }
  
}
