
package bg.unisofia.fmi.grosd.ns.util;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.List;

import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.exception.TaskLifecycleNotificationException;
import bg.unisofia.fmi.grosd.ns.listener.event.NSTaskAbortedEventImpl;
import bg.unisofia.fmi.grosd.ns.listener.visitor.NSTaskAbortedVisitor;
import bg.unisofia.fmi.grosd.ns.listener.visitor.NSTaskFinishedVisitor;
import bg.unisofia.fmi.grosd.ns.remote.listener.NSTaskListener;
import bg.unisofia.fmi.grosd.ns.remote.listener.visitor.NSListenerVisitor;
import bg.unisofia.fmi.grosd.ns.remote.task.NSTaskRemote;
import bg.unisofia.fmi.grosd.ns.remote.task.result.NSTaskResult;


/**
 * Utility class for notifying the registered task lifecycle listeners.
 *@author Shenol Yousouf
 */
public abstract class NSTaskLifecycleNotifier {
  
  /**
   * Notify about a failed task
   * @param task the original client's task, received from the API
   * @param errorMessage message of failure
   * @param cause list of <code>Exceptions</code> which are the cause for the error
   * @throws NodeServiceException if failed to notify the listeners
   */
  public static void notifyTaskAborted(NSTaskRemote task, String errorMessage, Exception[] cause) throws NodeServiceException {
    
    if (task == null) {
      // TODO: log error
      return;
    }
    
    List<NSTaskListener> taskListeners = null;
    try {
      taskListeners = task.getListeners();
    } catch (RemoteException e) {
      throw new TaskLifecycleNotificationException("Error while getting task listeners.", e);
    }
    
    if (taskListeners == null || taskListeners.size() == 0) {
      // no listeners to process
      return;
    }
    
    NSTaskAbortedEventImpl event = new NSTaskAbortedEventImpl(errorMessage, cause);
    NSTaskAbortedVisitor visitor = new NSTaskAbortedVisitor(event);
    NSListenerVisitor remoteVisitor = null;
    try {
      remoteVisitor = (NSListenerVisitor) UnicastRemoteObject.exportObject(visitor, 0);
    } catch (RemoteException e) {
      throw new TaskLifecycleNotificationException("Error while exporting NSTaskAbortedVisitor to remote object.",  e);
    }
    
    for (NSTaskListener taskListener:taskListeners) {
      try {
        taskListener.accept(remoteVisitor);
      } catch (RemoteException e) {
        throw new TaskLifecycleNotificationException("Error in notifying listeners for aborted task while accepting visitor.", e);
      }
    }
  }
  
  /**
   * Notify listeners about a finished taska and return the collected result.
   * @param task the original client's task, received from the API
   * @param result the generated <code>NSTaskResult</code> which encapsualtes the actual task result
   * @throws NodeServiceException if failed to notify the listeners
   */
  public static void notifyTaskFinished(NSTaskRemote task, NSTaskResult result) throws NodeServiceException {
    
    if (task == null || result == null) {
      // TODO: log error
      return;
    }
    
    List<NSTaskListener> taskListeners = null;
    try {
      taskListeners = task.getListeners();
    } catch (RemoteException e) {
      throw new TaskLifecycleNotificationException("Error while getting task listeners.", e);
    }
    
    if (taskListeners == null || taskListeners.size() == 0) {
      // no listeners to process
      return;
    }
    
    NSTaskFinishedVisitor visitor = new NSTaskFinishedVisitor(result);
    NSListenerVisitor remoteVisitor = null;
    try {
      remoteVisitor = (NSListenerVisitor) UnicastRemoteObject.exportObject(visitor, 0);
    } catch (RemoteException e) {
      throw new TaskLifecycleNotificationException("Error while exporting NSTaskFinishedVisitor to remote object.",  e);
    }
    
    for (NSTaskListener taskListener:taskListeners) {
      try {
        taskListener.accept(remoteVisitor);
      } catch (RemoteException e) {
        throw new TaskLifecycleNotificationException("Error in notifying listeners for finished task while accepting visitor.", e);
      }
    }
  }

}
