
package bg.unisofia.fmi.grosd.ns.util;

import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.ABORTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.FINISHED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STOPPED;

import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState;
import bg.unisofia.fmi.grosd.ns.impl.NodeServiceInternal;
import bg.unisofia.fmi.grosd.ns.remote.task.result.locator.NSFileFilter;
import bg.unisofia.fmi.grosd.ns.remote.task.result.locator.NSFileResultLocator;


/**
 * General-puspose class for performing some assisting tasks. 
 *@author Shenol Yousouf
 */
public abstract class NodeServiceUtilities {
  
  private final static Set<NSTaskActionState> finalStates = new HashSet<NSTaskActionState>();
  static {
    finalStates.add(ABORTED);
    finalStates.add(FINISHED);
    finalStates.add(STOPPED);
  }
  
  public final static String SERIALIZABLE_EXTENTION = ".ser";
  
  /**
   * The Exception stack trace is printed into a <code>String</code>.
   * @param ex exception
   * @return <code>String</code> representation of the exception
   */
  public static String formatExceptionMessage (Exception ex) {
    StringWriter strWr = new StringWriter();
    ex.printStackTrace(new PrintWriter(strWr));
    return "[" + ex.getClass().getName() + "] " + ex.getMessage() + "\n" + strWr.toString();
  }
  
  /**
   * 
   * @param taskID task identifier
   * @return the work folder into which the binary files of the task are transferred
   */
  public static File getWorkFolderForTask(String taskID) {
    return new File(NodeServiceInternal.getNodeServiceWorkFolder(), taskID);
  }
  
  /**
   * 
   * @param keyID identifier of the persistent task
   * @return the work folder into which the binary files of the task are transferred
   */
  public static File getWorkFolderForPersistentTask(String keyID) {
    return new File(NodeServiceInternal.getNSPersistentTasksFolder(), keyID);
  }
  
  /**
   * 
   * @param keyID key identifier of the persistent task
   * @return file storing the persistent task's metadata, required for its restoration
   */
  public static File getFileForPersistentTask(String keyID) {
    return new File(NodeServiceInternal.getNSPersistentTasksFolder(), keyID + SERIALIZABLE_EXTENTION);
  }
  
  /**
   * 
   * @param root root folder for searching
   * @param locator a <code>NSFileResultLocator</code> for marking the required files
   * @param comments comments, generated during search
   * @return all files that are marked from the locator
   * @throws RemoteException
   */
  public static File[] getOutputFiles (File root, NSFileResultLocator locator, StringBuilder comments) throws RemoteException {
    String[] folderNames = locator.getFolderNamesForSearch();
    NSFileFilter fileFilter = locator.getFileFilter();
    boolean inDepthSearch = locator.searchInSubfolders();
    
    ArrayList<File> results = new ArrayList<File>();
    
    if (folderNames == null || folderNames.length == 0) {
      // search in the root folder, from where the application is started
      results = getOutputFiles(root, fileFilter, inDepthSearch, results, comments);
    } else {
      for (String folderName : folderNames) {
        File file = new File (folderName);
        results = getOutputFiles(file, fileFilter, inDepthSearch, results, comments);
      }
    }
    
    return results.toArray(new File[] {});
  }
  
  private static ArrayList<File> getOutputFiles (File folder, NSFileFilter fileFilter, boolean inDepthSearch, ArrayList<File> results, StringBuilder comments) throws RemoteException {
    
    if (folder.exists() == false) {
      comments.append("File " + folder.getAbsolutePath() + " does not exists.");
      return results;
    }
    
    if (folder.isDirectory() == false) {
      comments.append("File " + folder.getAbsolutePath() + " is not a folder and will be skipped.");
      return results;
    }
    
    for (File f : folder.listFiles()) {
      if (fileFilter.accept(f)) {        // it is acceptable for the file filter to mark folders too; all of folder's contents will be copied too
        results.add(f);
      }
      
      if (f.isDirectory() && inDepthSearch) {
        results.addAll(getOutputFiles(f, fileFilter, inDepthSearch, results, comments));
      }
    }
    
    return results;
  }
  
  /**
   * 
   * @param state state for checking
   * @return if the checked state is final, i.e. one of the following: <code>FINISHED</code>, <code>ABORTED</code>, <code>STOPPED</code>
   */
  public static boolean isActionStateFinal(NSTaskActionState state) {
    return finalStates.contains(state);
  }
  
  /**
   * Delete file, allowing a certain timeout before giving up its deletion.
   * If the timeout is exceeded, because the file is locked by an open file handler,
   * the JVM is delegated to delete it on its exit, but this does not guarantee that the file will be actually deleted.
   * This method assumes that file can always be deleted, e.g. if file is folder, then the folder is empty, etc.
   * @param file file for deletion
   * @param timeout a period of time after which the file deletion is delegated to the JVM on its exit
   */
  public static void deleteFileWithTimeout (File file, long timeout) {
    
    System.out.println("Attempting to delete file " + file.getAbsolutePath());
    
    long startTime = System.currentTimeMillis();
    while (file.delete() == false && System.currentTimeMillis() - startTime < timeout) {
      try {
        Thread.sleep(100);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    }
    
    if (file.exists()) {
      System.out.println("Failed to delete file " + file.getAbsolutePath());
      file.deleteOnExit();                            // mark file for deletion on VM exit
    } else {
      System.out.println("File " + file.getAbsolutePath() + " deleted successfully.");      
    }
  }
}
