
package bg.unisofia.fmi.grosd.ns.util.thread;

import java.io.File;

import bg.unisofia.fmi.grosd.ns.enums.NodeServiceTaskType;
import bg.unisofia.fmi.grosd.ns.exception.NodeServiceException;
import bg.unisofia.fmi.grosd.ns.repository.NSRepositoryRecord;
import bg.unisofia.fmi.grosd.ns.repository.NSTaskRepository;


/**
 * Cleaner thread for removing teh files left after a task execution on a client's request.
 * The cleaning procedure in implemented in a separate thread to avoid the delays from
 * file operations and planned timeouts.
 *@author Shenol Yousouf
 */
public class NSTaskCleanerUtil extends Thread {
  
  private String taskID;
  private long fileTimeout;              // timeout for deleting a single file, defined in ms; if after a fileTimeout ms the file is still not deleted, the util gives up
  
  private NSRepositoryRecord taskRecord;
  
  public NSTaskCleanerUtil (NodeServiceTaskType taskType, String taskID, long timeout) throws NodeServiceException{
    this.taskID = taskID;
    this.fileTimeout = timeout;
    switch (taskType) {
    case NS_JAVA_GENERIC_TASK:
    case NS_JAVA_PERSISTENT_TASK_INVOCATION: this.taskRecord = NSTaskRepository.getTaskRecord(taskID);      
      break;
    case NS_JAVA_PERSISTENT_TASK_UPLOAD: this.taskRecord = NSTaskRepository.getServiceRecord(taskID);
      break;
    default: throw new NodeServiceException("The task type " + taskType.toString() + " is unknown or not supported for deleting garbage files.");
    }
    this.taskRecord = NSTaskRepository.getTaskRecord(taskID);
  }

  public void run() {
    
    File[] filesForDeletion = taskRecord.getFilesFromRecord();
    for (File file : filesForDeletion) {
      deleteFile(file);
    }
    // check if all files have been removed
    if (taskRecord.getFilesFromRecord().length == 0) {
      // no more need to hold data about task in repository
      NSTaskRepository.removeTaskRecord(taskID);
    }
  }
  
  private void deleteFile(File file) {
    if (file != null) {
      if (file.exists()) {
        File parentFile = file.getParentFile();
        if (file.isDirectory()) {
          for(File f : file.listFiles()) {
            deleteFile(f);
          }
          
          if (file.listFiles() == null || file.listFiles().length == 0) {       // empty folder
            deleteFileWithTimeout(file);
          } else {
            file.deleteOnExit();                      // mark file for deletion on VM exit
          }
        } else {
          deleteFileWithTimeout(file);
        }
        if (parentFile != null && parentFile.listFiles().length == 0) {
          deleteFileWithTimeout(parentFile);
        }
      } else {
        taskRecord.removeFilesFromRecord(file);
      }
    }
  }
  
  /**
   * This method assumes that file can always be deleted, e.g. if file is folder, then the folder is empty, etc.
   * @param file file for deletion
   */
  private void deleteFileWithTimeout (File file) {
    
    System.out.println("Attempting to delete file " + file.getAbsolutePath());
    
    long startTime = System.currentTimeMillis();
    while (file.delete() == false && System.currentTimeMillis() - startTime < this.fileTimeout) {
      try {
        Thread.sleep(100);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    }
    
    if (file.exists()) {
      System.out.println("Failed to delete file " + file.getAbsolutePath());
      file.deleteOnExit();                            // mark file for deletion on VM exit
    } else {
      System.out.println("File " + file.getAbsolutePath() + " deleted successfully.");
      taskRecord.removeFilesFromRecord(file);
    }
  }
}
