
package bg.unisofia.fmi.grosd.ns.api;

import java.rmi.RemoteException;

import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskInvocation;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskUpload;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaTask;


/**
 * Major Node Service API interface which acts as a direct client of Node Service kernel.
 * Used for establishing a connection to a certain node and for the upload of all suported sorts of tasks
 * to the Node Service kernel. Can be viewed like a proxy for the Node Service on a certain node
 * because all client requests pass through an instance of this interface first.
 *@author Shenol Yousouf
 */
public interface NodeServiceClient {

  /**
   * Upload a generic task for executing a public method, defined in a Java program.
   * @param task metadata objects which defines the task parameters 
   * @throws RemoteException
   */
  public void uploadJavaTask(NSJavaTask task) throws RemoteException;
  
  /**
   * Upload a persistent service for multiple invocation on the corresponding node.
   * After upload, the persistent service can be called for execution,
   * using <code>invokeJavaPersistentTask</code> method.
   * @param task    metadata objects which defines the task parameters
   * @throws RemoteException
   */
  public void uploadJavaPersistentTask(NSJavaPersistentTaskUpload task) throws RemoteException;
  
  /**
   * Invoke a persistent service, already uploaded to the node, with the supplied input data.
   * NOTE that once invoked, the service cannot process any other requests, until finished.
   * @param task metadata objects which defines the task parameters
   * @throws RemoteException
   */
  public void invokeJavaPersistentTask(NSJavaPersistentTaskInvocation task) throws RemoteException;
  
  /**
   * Stop an already uploaded task on the node before finishing execution.
   * @param taskID identifier for locating the task on the node
   * @param synchronous if <code>true</code>, this method will return with the actual task stop;
   * if <code>false</code>, the call will return as soon as possible, leaving the request to be processes asynchronously by the node
   * @throws RemoteException
   */
  public void stopJavaTask(final String taskID, final boolean synchronous) throws RemoteException;
  
  /**
   * Start a persistent task on the node. The task must be already present on the node before that, otherwise an exception is generated.
   * NOTE that you need start the task only if it has been stopped explicitly before that; the persistent services are automatically started after upload
   * and with the start of Node Service.
   * @param keyIdentifier   identifier for locating the task on the node
   * @throws RemoteException
   */
  public void startPersistentTask(String keyIdentifier) throws RemoteException;
  
  /**
   * Stop a persistent service on node. After stopping it, it cannot accept any requests for processing.
   * NOTE that after a Node Service restart, the task wil start again too.
   * @param keyIdentifier ID of the persistent task on node
   * @param forcedMode if true, the service will be stopped even in the middle of executing a request;
   * if false, the service will be stopped after the end of execution
   * @param synchronous if true, the method will end with the stop of the persistent task;
   * if false, the stopping is done asynchronously to the method call;
   * the synchronous parameter is introduced only here because stopping task  may take longe timer than starting it 
   * @throws RemoteException
   */
  public void stopPersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException;
  
  /**
   * Remove a persistent service completely from the node.
   * @param keyIdentifier ID of the persistent task on node
   * @param forcedMode if true, the service will be stopped even in the middle of executing a request;
   * if false, the service will be stopped after the end of execution
   * @param synchronous if true, the method will end with the stop of the persistent task;
   * if false, the stopping is done asynchronously to the method call;
   * the synchronous parameter is introduced only here because stopping task  may take longe timer than starting it 
   * @throws RemoteException
   */
  public void removePersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException;
  
  /**
   * Request removing of all created and transferred files that were used by this task.
   * The method is applicable only for tasks for one-time execution and for single invocations of persistent services;
   * to remove persistent services files, use <code>removePersistentTask</code> instead.
   * @param taskID identifier for locating the task on the node
   * @throws RemoteException
   */
  public void gcTask(String taskID) throws RemoteException;
  
  /**
   * @return name of the node host to which the current client is connected
   */
  public String getHostname();
  
  /**
   * @return port of the Node Service to which the current client is connected
   */
  public int getPort();
  
  /**
   * Removes some resources used by the inner connection between the current node and
   * the Node Service and after that the client cannot be used for uploading other requests to the node.
   */
  public void close();
}
