
package bg.unisofia.fmi.grosd.ns.api.task.args.impl;

import java.io.File;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

import bg.unisofia.fmi.grosd.ns.remote.task.arg.FileArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.NSTaskArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.SerializableArgument;


/**
 * Factory for creating NSTaskArguments from a various combinations of input data.
 *@author Shenol Yousouf
 */
public abstract class NSTaskArgumentFactory {

  /**************************************
  Create File Argument from provided File object
  **************************************/
  
  /**
   * Create a <code>FileArgument</code> wrapping the specified file.
   * The file will be copied to the working folder of the Node Service, by default.
   * @param local if <code>true</code>, this means that the file is not accessible for direct copy
   * and Node Service will utilize a special API callback object for its trarnsfer. 
   */
  public static FileArgument createFileArg (File file, boolean local) {
    try {
      return new FileArgumentImpl(file, local);
    } catch (MalformedURLException e) {
      throw new IllegalArgumentException("File path to file " + file.getName() + " cannot be parsed to URL.\nCause: " + e.getMessage());
    }
  }
  
  /**
   * Create a <code>FileArgument</code> wrapping the specified file.
   * The file will be copied to the specified relative destination on the Node Service host.
   * @param local if <code>true</code>, this means that the file is not accessible for direct copy
   * and Node Service will utilize a special API callback object for its trarnsfer. 
   */
  public static FileArgument createFileArg (File file, String destination, boolean local) {
    try {
      return new FileArgumentImpl(file, destination, local);
    } catch (MalformedURLException e) {
      throw new IllegalArgumentException("File path to file " + file.getName() + " cannot be parsed to URL.\nCause: " + e.getMessage());
    }
  }
  
  /**
   * Create a <code>FileArgument</code> wrapping the specified .class file.
   * The file will be copied to a destiantion, which complies to the provided classname.
   * @param local if <code>true</code>, this means that the file is not accessible for direct copy
   * and Node Service will utilize a special API callback object for its trarnsfer. 
   */
  public static FileArgument createClassFileArg(File file, String classname, boolean local) {
    return createFileArg(file, convertToDestiantionName(classname), local);
  }
  
  /**************************************
        Create File Argument from provided URL
  **************************************/
  
  /**
   * Same as <code>createFileArg (File file, boolean local)</code>
   * but file is specified with an URL instead.
   * @see NSTaskArgumentFactory.createFileArg (File file, boolean local)
   */
  public static FileArgument createFileArg(URL fileURL, boolean local) {
    try {
      return new FileArgumentImpl(fileURL, local);
    } catch (URISyntaxException e) {
      throw new IllegalArgumentException("Cannot instantiate File from URL " + fileURL);
    }
  }
  
  /**
   * Same as <code>createFileArg (File file, String destination, boolean local)</code>
   * but file is specified with an URL instead.
   * @see NSTaskArgumentFactory.createFileArg (File file, String destination, boolean local)
   */
  public static FileArgument createFileArg(URL fileURL, String destination, boolean local) {
    return new FileArgumentImpl(fileURL, destination, local);
  }
  
  /**
   * Same as <code>createFileArg (File file, String classname, boolean local)</code>
   * but file is specified with an URL instead.
   * @see NSTaskArgumentFactory.createFileArg (File file, String classname, boolean local)
   */
  public static FileArgument createClassFileArg(URL fileURL, String classname, boolean local) {
    return createFileArg(fileURL, convertToDestiantionName(classname), local);
  }
  
  /**************************************
                CreateSerializable Argument
**************************************/
  
  /**
   * Create a SerializableArgument, wrapping the specified <code>java.io.Serializable</code> object
   */
  public static SerializableArgument createSerializableArgument (Serializable ser) {
    return new SerializableArgumentImpl(ser);
  }
  
  /**
   * Create a SerializableArgument, wrapping the specified <code>java.io.Serializable</code> object,
   * specifying its proper classname too.
   */
  public static SerializableArgument createSerializableArgument (String formalParameterClassname, Serializable ser) {
    return new SerializableArgumentImpl(formalParameterClassname, ser);
  }
  
  /**************************************
                      Create Arguments List
**************************************/
  
  /**
   * Convert the specified list of objects into a list of <code> NSTaskArgument</code> objects in "one pass".
   * NOTE that this method accepts primitive types too, but the autoboxing feature of JDK5 wrapps them
   * into the corresponding wrapper classes. The Node Service provides some resolution features to
   * recognize the wrapped primitives but to avoid certain ambiguities, its is recommended to use
   * the proper <code>createSerializableArgument (String formalParameterClassname, Serializable ser)</code> method instead of this one.
   */
  public static NSTaskArgument[] createArgsList(Object...taskArguments ) {
    if (taskArguments == null) {
      return new NSTaskArgument[] {};
    }
    NSTaskArgument[] list = new NSTaskArgument[taskArguments.length];
    int index = 0;
    for (Object argument : taskArguments) {
      if (argument instanceof File) {
        list[index] = createFileArg((File)argument, true);
      } else if (argument instanceof FileArgument || argument instanceof SerializableArgument) {
        list[index] = (NSTaskArgument) argument;
      } else if (argument instanceof Serializable) {
        list[index] = createSerializableArgument((Serializable)argument);
      } else  {
        throw new IllegalArgumentException("Cannot build appropriate argument type from class " + argument.getClass().getName());
      }
      index++;
    }
    return list;
  }
  
  private static String convertToDestiantionName(String classname) {
    String destination = classname.trim().replace('.', File.separatorChar) + ".class";
    return destination;
  }
}
