
package bg.unisofia.fmi.grosd.ns.api.task.impl;

import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.ABORTED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.FINISHED;
import static bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState.STOPPED;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import bg.unisofia.fmi.grosd.ns.api.util.NodeServiceAPIUtilities;
import bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState;
import bg.unisofia.fmi.grosd.ns.remote.listener.NSActionStateListener;
import bg.unisofia.fmi.grosd.ns.remote.listener.NSTaskLifecycleListener;
import bg.unisofia.fmi.grosd.ns.remote.listener.NSTaskListener;
import bg.unisofia.fmi.grosd.ns.remote.listener.event.ActionStateChangedEvent;
import bg.unisofia.fmi.grosd.ns.remote.listener.visitor.NSListenerVisitor;
import bg.unisofia.fmi.grosd.ns.remote.listener.visitor.adapter.NSListenerVisitorAdapter;
import bg.unisofia.fmi.grosd.ns.remote.task.NSTaskRemote;


/**
 * An abstract implementation of <code>NSTaskRemote</code> which defines
 * all the common features that all tasks to support. All classes that intend to implement
 * the various task subtypes are recommended to inherit this class as well in order to
 * be consistent with other task types.
 * @see NSTaskRemote
 *@author Shenol Yousouf
 */
public abstract class NSAbstractTask implements NSTaskRemote {
  
  private String taskID;
  private NSTaskActionState actionState;
  private List< NSTaskListener> listeners = new ArrayList<NSTaskListener>();
  //private boolean exported = false;
  
  private static Set<NSTaskActionState> finalStates = new HashSet<NSTaskActionState>();
  static {
    finalStates.add(ABORTED);
    finalStates.add(FINISHED);
    finalStates.add(STOPPED);
  }
  
  protected NSAbstractTask() {
    this.taskID = NodeServiceAPIUtilities.generateTaskID();
    this.actionState = NSTaskActionState.UNSCHEDULED;
  }

  public String getTaskID() throws RemoteException {
    return taskID;
  }
  
  public NSTaskActionState getActionState() {
    return actionState;
  }
  
  
  public void addListener(NSTaskListener listener) throws RemoteException {
    listeners.add(listener);    
  }

  /* (non-Javadoc)
   * @see bg.unisofia.fmi.grosd.ns.task.NSTask#getListeners()
   */
  public List<NSTaskListener> getListeners() throws RemoteException {
    
    class AddListenerVisitor implements NSListenerVisitor {
      
      private List<NSTaskListener> remoteListeners;
      
      public AddListenerVisitor() {
        remoteListeners = new ArrayList<NSTaskListener>(listeners.size());
      }

      public void visit(NSActionStateListener listener) throws RemoteException {
        if (listener.isExported()) {
          remoteListeners.add(listener);
        } else {
          NSActionStateListener remoteListener = (NSActionStateListener) UnicastRemoteObject.exportObject(listener, 0);
          remoteListener.setExported();
          remoteListeners.add(remoteListener);
        }
      }
  
      public void visit(NSTaskLifecycleListener listener) throws RemoteException {
        if (listener.isExported()) {
          remoteListeners.add(listener);
        } else {
          NSTaskLifecycleListener remoteListener = (NSTaskLifecycleListener) UnicastRemoteObject.exportObject(listener, 0);
          remoteListener.setExported();
          remoteListeners.add(remoteListener);
        }
      }
      
      public List<NSTaskListener> getRemoteListeners() {
        return remoteListeners;
      }
    }
  

      AddListenerVisitor visitor = new AddListenerVisitor();
      for (NSTaskListener listener: listeners) {
        listener.accept(visitor);
      }    
      listeners = visitor.getRemoteListeners();
      return listeners;
    
  }
  /* (non-Javadoc)
   * @see bg.unisofia.fmi.grosd.ns.remote.task.NSTaskRemote#changeActionState(bg.unisofia.fmi.grosd.ns.enums.NSTaskActionState)
   */
  public void changeActionState(NSTaskActionState state, String comment) throws RemoteException {
    
    class ActionStateChangedEventImpl implements ActionStateChangedEvent {
  
        private NSTaskActionState oldState;
        private NSTaskActionState newState;
        private String comment;
      
        ActionStateChangedEventImpl (NSTaskActionState oldState, NSTaskActionState newState, String comment) {
          this.oldState = oldState;
          this.newState = newState;
          this.comment = comment;
        }
  
        public String getComment() {
          return comment;
        }
  
        public NSTaskActionState getNewActionState() {
          return newState;
        }
  
        public NSTaskActionState getOldActionState() {
          return oldState;
        }
      }
    
    class ActionStateChangedVisitor extends NSListenerVisitorAdapter {
      
      private ActionStateChangedEventImpl event;
      
       ActionStateChangedVisitor(ActionStateChangedEventImpl event) {
          this.event = event;
        }
        
        public void visit(NSActionStateListener listener) throws RemoteException {
            listener.actionStateChanged(event);
        }
      }
    
    if (finalStates.contains(actionState)) {
      throw new RemoteException("Illegal change of task state from " + actionState + " to " + state +
                          ".\nCurrent state " + actionState + " is final.");
    }
    
    NSTaskActionState oldState = actionState;
    this.actionState = state;
    
    ActionStateChangedEventImpl event = new ActionStateChangedEventImpl(oldState, state, comment);
    ActionStateChangedVisitor visitor = new ActionStateChangedVisitor(event);
    for (NSTaskListener listener : listeners) {
      listener.accept(visitor);
    }
  }  
}
