
package bg.unisofia.fmi.grosd.ns.api.task.impl;

import bg.unisofia.fmi.grosd.ns.api.task.args.impl.NSTaskArgumentFactory;
import bg.unisofia.fmi.grosd.ns.api.task.method.impl.NSTaskMethodFactory;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskInvocation;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaPersistentTaskUpload;
import bg.unisofia.fmi.grosd.ns.remote.task.NSJavaTask;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.FileArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.arg.NSTaskArgument;
import bg.unisofia.fmi.grosd.ns.remote.task.method.NSTaskMethod;
import bg.unisofia.fmi.grosd.ns.remote.task.method.NSTaskMethodPrototype;
import bg.unisofia.fmi.grosd.ns.remote.task.targetclass.NSTaskTargetClass;


/**
 * Factory for creating various subtask instances from different combinations of input data.
 *@author Shenol Yousouf
 */
public abstract class NSTaskFactory {

  /**************************************
      NSJavaTask Factory Methods
  **************************************/
  
  /**
   * Build task for one-time execuiton of a method of a Java program.
   * @param targetClass metadata for the class which is target of execution
   * @param taskMethod metadata for the method which is target of execution; the metadat also include the input argument values
   * @param binaries .jar and .class files which are required for the execuiton
   * @param additionalFiles additional files requried for the execution which are not included in the original input data of the task method
   */
  public static NSJavaTask createNSJavaTask(NSTaskTargetClass targetClass, NSTaskMethod taskMethod, FileArgument[] binaries, FileArgument[] additionalFiles ) { 
    return new NSJavaTaskImpl(targetClass, taskMethod, binaries, additionalFiles);
  }
  
  /**
   * Build task for one-time execuiton of a method of a Java program.
   * @param targetClass metadata for the class which is target of execution
   * @param taskMethod metadata for the method which is target of execution; the metadat also include the input argument values
   * @param binaries .jar and .class files which are required for the execuiton
   */
  public static NSJavaTask createNSJavaTask(NSTaskTargetClass targetClass, NSTaskMethod taskMethod, FileArgument[] binaries ) { 
    return new NSJavaTaskImpl(targetClass, taskMethod, binaries );
  }
  
  /**
   * Build task for one-time execuiton of a method of a Java program. As no binary files are specified for the task execution,
   * it is assumed that they were transferred and still present on the node after a previous task execution.
   * @param targetClass metadata for the class which is target of execution
   * @param taskMethod metadata for the method which is target of execution; the metadat also include the input argument values
   */
  public static NSJavaTask createNSJavaTask(NSTaskTargetClass targetClass, NSTaskMethod taskMethod ) { 
    return new NSJavaTaskImpl(targetClass, taskMethod);
  }
  
  /**************************************
      NSJavaMainTask Factory Methods
**************************************/
  
  /**
   * Build task for execution of a Java application's main-method.
   * @see createNSJavaTask method variants
   */
  public static NSJavaTask createNSJavaMainTask(NSTaskTargetClass targetClass, String[] mainArgs, FileArgument[] binaries, FileArgument[] additionalFiles) {
    NSTaskArgument[] nsArgs = NSTaskArgumentFactory.createArgsList((Object)mainArgs);
    NSTaskMethod mainMethod = NSTaskMethodFactory.createNSTaskMethod("main", nsArgs);
    return createNSJavaTask(targetClass, mainMethod, binaries, additionalFiles);
  }
  
  /**
   * Build task for execution of a Java application's main-method.
   * @see createNSJavaTask method variants
   */
  public static NSJavaTask createNSJavaMainTask(NSTaskTargetClass targetClass, String[] mainArgs, FileArgument[] binaries) {
    NSTaskArgument[] nsArgs = NSTaskArgumentFactory.createArgsList((Object)mainArgs);
    NSTaskMethod mainMethod = NSTaskMethodFactory.createNSTaskMethod("main", nsArgs);
    return createNSJavaTask(targetClass, mainMethod, binaries);
  }
  
  /**************************************
      NSJavaPersistenTaskUpload Factory Methods
**************************************/
  
  /**
   * Build a task for uploading a persistent service.
   * @param keyIdentifier identifier for addressing the invocation requests to; it has to be unique at least at a node level
   * @param targetClass class from which a method is executed on every invocation
   * @param binaries .jar and .class files which are required for the execuiton
   * @param defaultMethod default method for invocation
   */
  public static NSJavaPersistentTaskUpload createNSJavaPersistentTaskUpload (String keyIdentifier, NSTaskTargetClass targetClass, FileArgument[] binaries, NSTaskMethodPrototype defaultMethod) {
    return new NSJavaPersistentTaskUploadImpl(keyIdentifier, targetClass, binaries, defaultMethod);
  }
  
  /**
   * Build a task for uploading a persistent service. The method required for execution will be specified on every invocation request.
   * @param keyIdentifier identifier for addressing the invocation requests to; it has to be unique at least at a node level
   * @param targetClass class from which a method is executed on every invocation
   * @param binaries .jar and .class files which are required for the execuiton
   */
  public static NSJavaPersistentTaskUpload createNSJavaPersistentTaskUpload(String keyIdentifier, NSTaskTargetClass targetClass, FileArgument[] binaries) {
    return new NSJavaPersistentTaskUploadImpl(keyIdentifier, targetClass, binaries);
  }
  
  /**************************************
  NSJavaPersistenTaskInvocation Factory Methods
**************************************/
  
  /**
   * Build task for invoking an already uploaded persistent task corresponding to the key identifier
   * by specifiying method and initailization parameters that
   * are different than the default ones with which the persistent task has been uploaded.
   * The default method and initialization paramters replacement is valid for this invocation only.
   * The unwrapped argument values are supplied to the specified method for invocation.
   */
  public static NSJavaPersistentTaskInvocation createNSJavaPersistentTaskInvocation (String keyIdentifier, NSTaskArgument[] args, FileArgument[] files,  NSTaskMethodPrototype method, String[] initClassnames, NSTaskArgument[] initArgs) {
    return new NSJavaPersistentTaskInvocationImpl (keyIdentifier, args, files, method, initClassnames, initArgs);
  }
  
  /**
  * Build task for invoking an already uploaded persistent task corresponding to the key identifier
   * by specifiying initailization parameters that
  * are different than the default ones with which the persistent task has been uploaded.
  * The default  initialization paramters replacement is valid for this invocation only.
  * The unwrapped argument values are supplied to the persistent service's method for invocation.
  */
  public static NSJavaPersistentTaskInvocation createNSJavaPersistentTaskInvocation (String keyIdentifier, NSTaskArgument[] args, FileArgument[] files, String[] initClassnames, NSTaskArgument[] initArgs) {
    return new NSJavaPersistentTaskInvocationImpl (keyIdentifier, args, files, null, initClassnames, initArgs);
  }
  
  /**
  * Build task for invoking an already uploaded persistent task corresponding to the key identifier
   *by specifiying method that
  * is different than the default one with which the persistent task has been uploaded.
  * The default method replacement is valid for this invocation only.
  * The unwrapped argument values are supplied to the specified method for invocation.
  */
  public static NSJavaPersistentTaskInvocation createNSJavaPersistentTaskInvocation (String keyIdentifier, NSTaskArgument[] args, FileArgument[] files,  NSTaskMethodPrototype method) {
    return new NSJavaPersistentTaskInvocationImpl (keyIdentifier, args, files, method);
  }
  
  /**
   * Build task for invoking an already uploaded persistent task  corresponding to the key identifier
   * by specifiying its input parameters and additonally needed input files.
   * The unwrapped argument values are supplied to the persistent service's method for invocation.
   */
  public static NSJavaPersistentTaskInvocation createNSJavaPersistentTaskInvocation (String keyIdentifier, NSTaskArgument[] args, FileArgument[] files) {
    return new NSJavaPersistentTaskInvocationImpl (keyIdentifier, args, files);
  }
  
  /**
   * Build task for invoking an already uploaded persistent task  corresponding to the key identifier
   * by specifiying its input parameters.
   * The unwrapped argument values are supplied to the persistent service's method for invocation.
   */
  public static NSJavaPersistentTaskInvocation createNSJavaPersistentTaskInvocation (String keyIdentifier, NSTaskArgument[] args) {
    return new NSJavaPersistentTaskInvocationImpl (keyIdentifier, args);
  }
}
