
package bg.unisofia.fmi.grosd.ns.remote;

import java.rmi.Remote;
import java.rmi.RemoteException;

import bg.unisofia.fmi.grosd.ns.enums.NodeServiceTaskType;
import bg.unisofia.fmi.grosd.ns.remote.task.NSTaskRemote;


/**
 * Central interface of Node Service working on a node.
 * Defines the set of services it can render to the Node Service API.
 *@author Shenol Yousouf
 */
public interface NodeService extends Remote {
  
  // binding name for the Node Service in the RMI registry
  final String NodeServiceName = "NodeService";
  // default port on which the Node Service is registered as a RMI service
  final int NodeServicePort = 2007;
  
  /**
   * Upload generic task for execution on node 
   * @param taskType type for proper task idenitification on node
   * @param task    task for execution
   * @param client  client object for callback
   * @throws RemoteException
   */
  public void uploadTask(NodeServiceTaskType taskType, NSTaskRemote task, NodeServiceClientCallback client) throws RemoteException;
  
  /**
   * Remove all resources left from the execution of an already finished task on node.
   * @param taskID  ID for identifiying the task on the node
   * @throws RemoteException
   */
  public void gcTaskResources(String taskID) throws RemoteException;
  
  /**
   * Stop an already uploaded Java task from executing on node
   * @param taskID  ID for identifiying the task on the node
   * @param synchronous if <code>true</code>, the method will return after the task has been actually stopped; if <code>false</code>, the method will return as soon as possible
   * @throws RemoteException
   */
  public void stopJavaTask(final String taskID, final boolean synchronous) throws RemoteException;
  
  /**
   * Start persistent service, already uploaded on a node.
   * Note that you need to call this method only if you have stopped  the service explicitly before that;
   * the service starts automatically after upload or with the Node Service start. 
   * @param keyIdentifier  persistent task of  identifier on node 
   * @throws RemoteException
   */
  public void startPersistentTask(String keyIdentifier) throws RemoteException;
  
  /**
   * Stop persistent task, upoloaded on a node.
   * @param keyIdentifier ID of the persistent task on node
   * @param forcedMode if true, the service will be stopped even in the middle of executing a request;
   * if false, the service will be stopped after the end of execution
   * @param synchronous if <code>true</code>, the method will return with the stop of the persistent task;
   * if <code>false</code>, the stopping is done asynchronously to the method call;
   * the synchronous parameter is introduced only on stopping operations because stopping task  may take longer time than starting it 
   * @throws RemoteException
   */
  public void stopPersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException;
  
  /**
   * Remove persistent service, uploaded on node.
   * @param keyIdentifier ID of the persistent task on node
   * @param forcedMode if <code>true</code>, the service will be stopped even in the middle of executing a request;
   * if <code>false</code>, the service will be stopped after the end of execution
   * @param synchronous if <code>true</code>, the method will return with the stop of the persistent task;
   * if <code>false</code>, the stopping is done asynchronously to the method call;
   * the synchronous parameter is introduced only on stopping operations because stopping task  may take longer time than starting it 
   * @throws RemoteException
   */
  public void removePersistentTask(final String keyIdentifier, final boolean forcedMode, final boolean synchronous) throws RemoteException;
  
  
  /**
   * Stop the Node Service.
   * @param forcedMode if true, the service will be stopped even in the middle of executing a request;
   * if false, the service will wait for the end of execution before stopping
   * @param synchronous if true, the method will end with the stop of the Node Service;
   * if false, the stopping is done asynchronously to the method call;
   * the synchronous parameter is introduced only on stopping operations because stopping a task may take longe timer than starting it 
   * @throws RemoteException
   */
  // in the current design, it is projected only the Node Service Server to be able to stop the NS (implementation and tests are not done yet)
  // if it is considered that this option should be extended to the NS API as well, uncomment the line below
  //public void stopNodeService(final boolean forcedMode, final boolean synchronous) throws RemoteException;
}
