
package bg.unisofia.fmi.grosd.ns.remote.task.result;

import java.io.Serializable;
import java.net.URL;
import java.rmi.Remote;
import java.rmi.RemoteException;


/**
 * Interface for getiing the result from executing the task on the node.
 * The interface declares numerous methods for getting various types of results but note that
 * usually in most cases not all of them will return a result if the return type is not relevant to the actual result type.
 * Node Service supports execution of tasks that can return a result from its target method and writes ouptut files
 * at the same time. In this case, if the proper file locator has been defined with the task, this interface will be able to
 * return both types of results.
 * The basic groups of returned result types that this interface supports are:
 * <li>as a <code>java.io.Serializable</code> instance</li> 
 * <li>as a primitive argument</li>
 * <li>as a list of file URLs for direct access on NS API host</li>
 *@author Shenol Yousouf
 */
public interface NSTaskResult extends Remote, Serializable {

  /**
   * @return result as <code>java.io.Serializable</code> object. Note that the result cannot be obtained simply as
   * a <code>java.lang.Object</code>, because <code>Objects</code> cannot be directly transferred between hosts as arguments of RMI calls.
   * @throws RemoteException
   */
  public Serializable getResult() throws RemoteException;                          // the result must be in a serializable form for the transfer
  
  /**
   * @return result as <code>boolean</code>
   * @throws RemoteException
   */
  public boolean getReultAsBoolean() throws RemoteException;               // who is gonna need this !!! :)
  /**
   * @return result as <code>byte</code>
   * @throws RemoteException
   */
  public byte getResultAsByte() throws RemoteException;
  /**
   * @return result as <code>char</code>
   * @throws RemoteException
   */
  public char getResultAsChar() throws RemoteException;
  /**
   * @return result as <code>short</code>
   * @throws RemoteException
   */
  public short getResultAsShort() throws RemoteException;
  /**
   * @return result as <code>int</code>
   * @throws RemoteException
   */
  public int getResultAsInt() throws RemoteException;
  /**
   * @return result as <code>long</code>
   * @throws RemoteException
   */
  public long getResultAsLong() throws RemoteException;
  /**
   * @return result as <code>float</code>
   * @throws RemoteException
   */
  public float getResultAsFloat() throws RemoteException;
  /**
   * @return result as <code>double</code>
   * @throws RemoteException
   */
  public double getResultAsDouble() throws RemoteException;
  
  /**
   * @return result as a list of local URLs to output files, generated during task execution;
   * NOTE that this is the method with processing logic and can be executed slowly -
   * on its invocation the files from the Node Service host are copied to the NS API host
   * and then URLs to these files are returned for direct local access.
   * @throws RemoteException
   */
  public URL[] getResultAsFileURLs() throws RemoteException;
  
  /**
   * @return some comment about the collected result
   * @throws RemoteException
   */
  public String getComment() throws RemoteException;
}
